# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import json
import socket
import time
import urllib.error
import urllib.parse
import urllib.request

from trackma import utils
from trackma.lib.lib import lib


class libshikimori(lib):
    """
    API class to communicate with Shikimori

    Website: https://shikimori.one

    messenger: Messenger object to send useful messages to
    """
    name = 'libshikimori'
    msg = None
    logged_in = False

    api_info = {'name': 'Shikimori', 'shortname': 'shikimori',
                'version': 2, 'merge': False}
    mediatypes = dict()
    mediatypes['anime'] = {
        'has_progress': True,
        'can_add': True,
        'can_delete': True,
        'can_score': True,
        'can_status': True,
        'can_update': True,
        'can_play': True,
        'statuses_start': ['watching', 'rewatching'],
        'statuses_finish': ['completed'],
        'statuses_library': ['watching', 'rewatching', 'planned'],
        'statuses':  ['watching', 'completed', 'on_hold', 'rewatching', 'dropped', 'planned'],
        'statuses_dict': {
            'watching': 'Watching',
            'completed': 'Completed',
            'on_hold': 'On-Hold',
            'rewatching': 'Rewatching',
            'dropped': 'Dropped',
            'planned': 'Plan to Watch'
        },
        'score_max': 10,
        'score_step': 1,
    }
    mediatypes['manga'] = {
        'has_progress': True,
        'can_add': True,
        'can_delete': True,
        'can_score': True,
        'can_status': True,
        'can_update': True,
        'can_play': False,
        'statuses_start': ['watching', 'rewatching'],
        'statuses_finish': ['completed'],
        'statuses':  ['watching', 'completed', 'on_hold', 'rewatching', 'dropped', 'planned'],
        'statuses_dict': {
            'watching': 'Reading',
            'completed': 'Completed',
            'on_hold': 'On-Hold',
            'rewatching': 'Re-reading',
            'dropped': 'Dropped',
            'planned': 'Plan to Read'
        },
        'score_max': 10,
        'score_step': 1,
    }
    default_mediatype = 'anime'

    # Supported signals for the data handler
    signals = {'show_info_changed': None, }

    url = "https://shikimori.one"
    auth_url = "https://shikimori.one/oauth/token"
    api_url = "https://shikimori.one/api"

    client_id = "Jfu9MKkUKPG4fOC95A6uwUVLHy3pwMo3jJB7YLSp7Ro"
    client_secret = "y7YmQx8n1l7eBRugUSiB7NfNJxaNBMvwppfxJLormXU"

    status_translate = {
        'ongoing': utils.Status.AIRING,
        'released': utils.Status.FINISHED,
        'anons': utils.Status.NOTYET,
        'cancelled': utils.Status.CANCELLED,
    }

    type_translate = {
        None: utils.Type.UNKNOWN,
        'tv': utils.Type.TV,
        'movie': utils.Type.MOVIE,
        'ova': utils.Type.OVA,
        'ona': utils.Type.ONA,
        'special': utils.Type.SP,
        'music': utils.Type.OTHER,
        'tv_13': utils.Type.TV,
        'tv_24': utils.Type.TV,
        'tv_48': utils.Type.TV,
    }

    def __init__(self, messenger, account, userconfig):
        """Initializes the API"""
        super(libshikimori, self).__init__(messenger, account, userconfig)

        self.pin = account['password']
        self.userid = self._get_userconfig('userid')

        if not self.pin:
            raise utils.APIFatal("No PIN.")

        if self.mediatype == 'manga':
            self.total_str = "chapters"
            self.watched_str = "chapters"

        else:
            self.total_str = "episodes"
            self.watched_str = "episodes"

        # handler=urllib.request.HTTPHandler(debuglevel=1)
        # self.opener = urllib.request.build_opener(handler)
        self.opener = urllib.request.build_opener()
        self.opener.addheaders = [('User-agent', 'Trackma')]

    def _request(self, method, url, get=None, post=None, jsondata=None, auth=False):
        content_type = None

        if get:
            url += "?%s" % urllib.parse.urlencode(get)
        if post:
            post = urllib.parse.urlencode(post).encode('utf-8')
            content_type = 'application/x-www-form-urlencoded'
        if jsondata:
            post = json.dumps(jsondata).encode('utf-8')
            content_type = 'application/json'

        request = urllib.request.Request(url, post)
        self.msg.debug("URL: %s" % url)
        request.get_method = lambda: method

        if content_type:
            request.add_header('Content-Type', content_type)

        if auth:
            request.add_header('Authorization', '{0} {1}'.format(
                self._get_userconfig('token_type').capitalize(),
                self._get_userconfig('access_token'),
            ))

        try:
            response = self.opener.open(request)

            return json.loads(response.read().decode('utf-8'))
        except urllib.error.URLError as e:
            raise utils.APIError("URL error: %s" % e)
        except socket.timeout:
            raise utils.APIError("Operation timed out.")

    def _request_access_token(self, refresh=False):
        """
        Requests or refreshes the access token through OAuth2
        """
        params = {
            'client_id':     self.client_id,
            'client_secret': self.client_secret,
            'redirect_uri':  'urn:ietf:wg:oauth:2.0:oob',
        }

        if refresh:
            self.msg.info('Refreshing access token...')

            params['grant_type'] = 'refresh_token'
            params['refresh_token'] = self._get_userconfig('refresh_token')
        else:
            self.msg.info('Requesting access token...')

            params['code'] = self.pin
            params['grant_type'] = 'authorization_code'

        data = self._request('POST', self.auth_url, post=params)

        timestamp = int(time.time())

        self._set_userconfig('access_token',  data['access_token'])
        self._set_userconfig('token_type',    data['token_type'])
        self._set_userconfig('expires',       timestamp + data['expires_in'])
        self._set_userconfig('refresh_token', data['refresh_token'])

        self.logged_in = True
        self._emit_signal('userconfig_changed')

    def _refresh_user_info(self):
        self.msg.info('Refreshing user details...')

        data = self._request("GET", self.api_url + "/users/whoami", auth=True)

        self._set_userconfig('userid', data['id'])
        self._set_userconfig('username', data['nickname'])

        self.userid = data['id']
        self._emit_signal('userconfig_changed')

    def check_credentials(self):
        timestamp = int(time.time())

        if not self._get_userconfig('access_token'):
            self._request_access_token(False)
        elif (timestamp+60) > self._get_userconfig('expires'):
            self._request_access_token(True)
        else:
            self.logged_in = True

        if not self.userid:
            self._refresh_user_info()

        return True

    def fetch_list(self):
        self.check_credentials()
        self.msg.info('Downloading list...')

        params = {'limit': 5000}
        data = self._request(
            "GET", self.api_url + "/users/{}/{}_rates".format(self.userid, self.mediatype), get=params)

        showlist = {}
        for item in data:
            show = utils.show()
            showid = item[self.mediatype]['id']
            show.update({
                'id': showid,
                'my_id': item['id'],
                'title': item[self.mediatype]['name'],
                'aliases': [item[self.mediatype]['russian']],
                'type': self.type_translate[item[self.mediatype]['kind']],
                'status': self.status_translate[item[self.mediatype]['status']],
                'my_progress': item[self.watched_str],
                'my_status': item['status'],
                'my_score': item['score'],
                'total': item[self.mediatype][self.total_str],
                'url': self.url + item[self.mediatype]['url'],
                'image': self.url + item[self.mediatype]['image']['original'],
                'image_thumb': self.url + item[self.mediatype]['image']['preview'],
            })

            showlist[showid] = show

        return showlist

    def add_show(self, item):
        self.check_credentials()
        self.msg.info("Adding item %s..." % item['title'])
        return self._update_entry(item, "POST")

    def update_show(self, item):
        self.check_credentials()
        self.msg.info("Updating item %s..." % item['title'])
        return self._update_entry(item, "PUT")

    def delete_show(self, item):
        self.check_credentials()
        self.msg.info("Deleting item %s..." % item['title'])

        data = self._request(
            "DELETE", self.api_url + "/user_rates/{}".format(item['my_id']), auth=True)

    def search(self, criteria, method):
        self.check_credentials()

        self.msg.info("Searching for {}...".format(criteria))
        param = {'q': criteria}
        try:
            data = self._request(
                "GET", self.api_url + "/{}s/search".format(self.mediatype), get=param)
        except ValueError:
            # An empty document, without any JSON, is returned
            # when there are no results.
            return []

        showlist = []

        for item in data:
            show = utils.show()
            showid = item['id']
            show.update({
                'id': showid,
                'title': item['name'],
                'aliases': [item['russian']],
                'type': self.type_translate[item['kind']],
                'status': self.status_translate[item['status']],
                'my_status': self.media_info()['statuses_start'][0],
                'total': item[self.total_str],
                'image': self.url + item['image']['original'],
                'image_thumb': self.url + item['image']['preview'],
            })

            showlist.append(show)

        return showlist

    def request_info(self, itemlist):
        self.check_credentials()
        infolist = []

        for show in itemlist:
            data = self._request(
                "GET", self.api_url + "/{}s/{}".format(self.mediatype, show['id']))
            infolist.append(self._parse_info(data))

        self._emit_signal('show_info_changed', infolist)
        return infolist

    def media_info(self):
        """Return information about the currently selected mediatype."""
        return self.mediatypes[self.mediatype]

    def _update_entry(self, item, method):
        # Note: This method returns the newly added or modified item ID (my_id)
        if method == 'POST':
            user_rate = {
                'user_id': self.userid,
                'target_id': item['id'],
                'target_type': self.mediatype.capitalize(),
            }
            dest_url = self.api_url + "/user_rates"
        else:
            # user_rate = {'score': 0, 'status': 0, 'episodes': 0, 'volumes': 0, 'chapters': 0, 'text': '', 'rewatches': 0}
            user_rate = {}
            dest_url = self.api_url + "/user_rates/{}".format(item['my_id'])

        if 'my_progress' in item:
            user_rate[self.watched_str] = item['my_progress']
        if 'my_status' in item:
            user_rate['status'] = item['my_status']
        if 'my_score' in item:
            user_rate['score'] = item['my_score']

        values = {'user_rate': user_rate}
        data = self._request(method, dest_url, jsondata=values, auth=True)
        return data['id']

    def _parse_info(self, item):
        info = utils.show()
        info.update({
            'id': item['id'],
            'title': item['name'],
            'type': self.type_translate[item['kind']],
            'status': self.status_translate[item['status']],
            'image': self.url + item['image']['original'],
            'url': self.url + item['url'],
            'extra': [
                ('Description',     self._lc(item.get('description'))),
                # ('Genres',          item.get('genres')),
                ('Type',            self._lc(item.get('kind').capitalize())),
                ('Average score',   self._lc(item.get('score'))),
                ('Russian title',   self._lc(item.get('russian'))),
                ('Japanese title',  self._lc(item.get('japanese')[0])),
                ('English title',   self._lc(item.get('english'))),
            ]
        })
        return info

    def _lc(self, v):
        if v == [None]:
            return None
        return v



import os
import queue

from PyQt6 import QtCore, QtGui, QtNetwork

ATTRIB_FILE = QtNetwork.QNetworkRequest.Attribute(1000)
ATTRIB_ID = QtNetwork.QNetworkRequest.Attribute(1001)


class ThumbManager(QtCore.QObject):
    """
    This class manages the async downloads of thumbnails.
    getThumb gets a thumbnail from the cache, and
    queueDownload starts a remote download.

    When a remote download is finished, the itemFinished signal
    is emitted, along with the ID and a QImage object.
    """

    itemFinished = QtCore.pyqtSignal(str, QtGui.QImage)

    def __init__(self, parent=None):
        self.manager = QtNetwork.QNetworkAccessManager()
        self.queue = queue.Queue()
        self.downloads = {}

        super().__init__(parent)

    def exists(self, filename):
        return os.path.isfile(filename)

    def getThumb(self, filename):
        return QtGui.QImage(filename)

    def queueDownload(self, iid, url, filename):
        self.queue.put((iid, url, filename))
        self.processNext()

    def processNext(self):
        if not self.queue.empty() and len(self.downloads) < 3:
            (iid, url, filename) = self.queue.get()

            request = QtNetwork.QNetworkRequest(QtCore.QUrl(url))
            request.setAttribute(
                QtNetwork.QNetworkRequest.Attribute(ATTRIB_ID), iid)
            request.setAttribute(
                QtNetwork.QNetworkRequest.Attribute(ATTRIB_FILE), filename)

            reply = self.manager.get(request)
            self.downloads[filename] = reply

            reply.finished.connect(self.onItemFinished)

    def onItemFinished(self):
        reply = self.sender()
        iid = str(reply.request().attribute(ATTRIB_ID))
        fname = reply.request().attribute(ATTRIB_FILE)

        data = reply.readAll()
        image = QtGui.QImage.fromData(data)
        thumb = image.scaled(
            200, 280, QtCore.Qt.AspectRatioMode.KeepAspectRatio, QtCore.Qt.TransformationMode.SmoothTransformation)
        thumb.save(fname)

        self.downloads.pop(fname)
        self.itemFinished.emit(iid, thumb)
        self.processNext()



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import GObject


class ShowEventType(GObject.GEnum):
    NONE = 0
    DETAILS = 1
    OPEN_WEBSITE = 2
    OPEN_FOLDER = 3
    COPY_TITLE = 4
    CHANGE_ALTERNATIVE_TITLE = 5
    REMOVE = 6
    PLAY_NEXT = 7
    PLAY_EPISODE = 8
    PLAY_RANDOM = 9
    EPISODE_ADD = 10
    EPISODE_SET = 11
    EPISODE_REMOVE = 12
    SET_SCORE = 13
    SET_STATUS = 14



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import webbrowser
from enum import Enum

from gi.repository import GObject, GdkPixbuf, Gtk

from trackma import utils
from trackma.ui.gtk import gtk_dir


class AccountsView(Enum):
    LIST = 1
    NEW = 2
    EDIT = 3


@Gtk.Template.from_file(os.path.join(gtk_dir, 'data/accountswindow.ui'))
class AccountsWindow(Gtk.Dialog):

    __gtype_name__ = 'AccountsWindow'

    __gsignals__ = {
        'account-cancel': (GObject.SignalFlags.RUN_FIRST, None, ()),
        'account-open': (GObject.SignalFlags.RUN_FIRST, None, (int, bool))
    }

    internal_box = Gtk.Template.Child()
    accounts_frame = Gtk.Template.Child()
    accounts_listbox = Gtk.Template.Child()
    revealer_action_bar = Gtk.Template.Child()
    btn_cancel = Gtk.Template.Child()
    btn_add = Gtk.Template.Child()
    btn_new_confirm = Gtk.Template.Child()
    btn_new_cancel = Gtk.Template.Child()
    btn_edit_confirm = Gtk.Template.Child()
    remember_switch = Gtk.Template.Child()
    accounts_stack = Gtk.Template.Child()
    accounts_combo = Gtk.Template.Child()
    password_label = Gtk.Template.Child()
    password_entry = Gtk.Template.Child()
    username_label = Gtk.Template.Child()
    username_entry = Gtk.Template.Child()
    btn_pin_request = Gtk.Template.Child()

    def __init__(self, manager, transient_for=None):
        Gtk.Dialog.__init__(self, use_header_bar=True,
                            transient_for=transient_for)
        self.init_template()

        self.accounts = []
        self.pixbufs = {}
        self.treeiters = {}
        self.current = AccountsView.LIST
        self.manager = manager
        self.account_edit = None

        self.adding_allow = False
        self.adding_extra = {}

        self._remove_border()
        self._add_separators()
        self._refresh_remember()
        self._refresh_pixbufs()
        self._refresh_list()
        self._populate_combobox()

    def _remove_border(self):
        self.internal_box.set_border_width(0)

    def _add_separators(self):
        self.accounts_listbox.set_header_func(
            self._accounts_listbox_header_func, None)

    @staticmethod
    def _accounts_listbox_header_func(row, before, _user_data):
        if before is None:
            row.set_header(None)
            return

        current = row.get_header()
        if current is None:
            current = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
            current.show()
            row.set_header(current)

    def _refresh_remember(self):
        remember = self.manager.get_default() is not None
        self.remember_switch.set_active(remember)

    def _refresh_pixbufs(self):
        self.pixbufs = {}
        for (libname, lib) in utils.available_libs.items():
            self.pixbufs[libname] = GdkPixbuf.Pixbuf.new_from_file(lib[1])

    def _refresh_list(self):
        for account in self.accounts:
            account.destroy()

        self.accounts = []

        for k, account in self.manager.get_accounts():
            libname = account['api']
            try:
                api = utils.available_libs[libname]
                account = AccountRow({
                    'number': k,
                    'username': account['username'],
                    'libname': libname,
                    'libname_desc': api[0],
                    'logo': self.pixbufs[libname],
                    'active': True
                })

            except KeyError:
                # Invalid API
                account = AccountRow({
                    'number': k,
                    'username': account['username'],
                    'libname': None,
                    'libname_desc': 'N/A',
                    'logo': None,
                    'active': False
                })

            self.accounts_listbox.add(account)
            self.accounts.append(account)

        if not self.accounts:
            self.accounts_frame.hide()
        else:
            self.accounts_frame.show()

    @Gtk.Template.Callback()
    def _on_dialog_close(self, dialog):
        self.emit('account-cancel')
        self.destroy()

    @Gtk.Template.Callback()
    def _on_btn_cancel_clicked(self, btn):
        self.emit('account-cancel')
        self.destroy()

    @Gtk.Template.Callback()
    def _on_row_selected(self, list_box, row):
        reveal = row is not None
        self.revealer_action_bar.set_reveal_child(reveal)

    @Gtk.Template.Callback()
    def _on_row_activated(self, list_box, row):
        acc_num = row.get_account_id()
        remember = self.remember_switch.get_active()
        self.emit('account-open', acc_num, remember)
        self.destroy()

    @Gtk.Template.Callback()
    def _on_btn_edit_clicked(self, btn):
        self._show_edit()

    @Gtk.Template.Callback()
    def _on_btn_open_clicked(self, btn):
        acc_num = self.accounts_listbox.get_selected_row().get_account_id()
        remember = self.remember_switch.get_active()
        self.emit('account-open', acc_num, remember)
        self.destroy()

    @Gtk.Template.Callback()
    def _on_btn_delete_clicked(self, btn):
        row = self.accounts_listbox.get_selected_row()
        self.manager.delete_account(row.get_account_id())
        row.destroy()

    @Gtk.Template.Callback()
    def _on_btn_add_clicked(self, btn):
        self._show_add_new()

    @Gtk.Template.Callback()
    def _on_btn_new_cancel_clicked(self, btn):
        self._show_accounts_list()

    @Gtk.Template.Callback()
    def _on_btn_new_confirm_clicked(self, btn):
        self._add_account()
        self._refresh_list()
        self._show_accounts_list()

    @Gtk.Template.Callback()
    def _on_btn_edit_confirm_clicked(self, btn):
        self._edit_account()
        self._refresh_list()
        self._show_accounts_list()

    def _show_edit(self):
        row = self.accounts_listbox.get_selected_row()
        self.account_edit = self.manager.get_account(row.get_account_id())
        self.account_edit['account_id'] = row.get_account_id()

        self.set_title("Edit account")
        self._clear_new_account()

        if utils.available_libs[self.account_edit['api']][2] == utils.Login.OAUTH:
            self._show_oauth_account()
        else:
            self._show_password_account()

        self.accounts_combo.set_active_iter(self.treeiters[row.get_libname()])
        self.username_entry.set_text(self.account_edit['username'])
        self.password_entry.set_text(self.account_edit['password'])

        self.accounts_combo.set_sensitive(False)
        self.username_entry.set_sensitive(False)

        self.btn_new_confirm.hide()
        self.btn_new_cancel.show()
        self.btn_edit_confirm.show()
        self.btn_add.hide()
        self.btn_cancel.hide()

        self.current = AccountsView.EDIT
        self.accounts_stack.set_visible_child_full(
            'new_account', Gtk.StackTransitionType.SLIDE_LEFT)

    def _show_add_new(self):
        self.set_title("Add account")
        self._clear_new_account()
        self.btn_new_confirm.show()
        self.btn_new_cancel.show()
        self.btn_add.hide()
        self.btn_cancel.hide()
        self.accounts_combo.set_sensitive(True)
        self.username_entry.set_sensitive(True)

        self.current = AccountsView.NEW
        self.accounts_stack.set_visible_child_full(
            'new_account', Gtk.StackTransitionType.SLIDE_LEFT)

    def _show_accounts_list(self):
        self.set_title("Accounts")
        self.btn_new_confirm.hide()
        self.btn_new_cancel.hide()
        self.btn_edit_confirm.hide()
        self.btn_add.show()
        self.btn_cancel.show()

        self.current = AccountsView.LIST
        self.accounts_stack.set_visible_child_full(
            'accounts', Gtk.StackTransitionType.SLIDE_RIGHT)

    def _populate_combobox(self):
        model_api = Gtk.ListStore(str, str, GdkPixbuf.Pixbuf)

        for (libname, lib) in sorted(utils.available_libs.items()):
            self.treeiters[libname] = model_api.append(
                [libname, lib[0], self.pixbufs[libname]])

        self.accounts_combo.set_model(model_api)

    @Gtk.Template.Callback()
    def _on_accounts_combo_changed(self, combo):
        self.username_entry.set_text("")
        self.password_entry.set_text("")
        api = self._get_combo_active_api_name()

        if not api or utils.available_libs[api][2] in [utils.Login.OAUTH, utils.Login.OAUTH_PKCE]:
            # We'll only allow adding the account if the user requests the PIN
            self.adding_allow = False
            self._show_oauth_account()
        else:
            self.adding_allow = True
            self._show_password_account()

    @Gtk.Template.Callback()
    def _on_btn_pin_request_clicked(self, btn):
        api = self._get_combo_active_api_name()

        auth_url = utils.available_libs[api][3]
        if utils.available_libs[api][2] == utils.Login.OAUTH_PKCE:
            self.adding_extra = {'code_verifier': utils.oauth_generate_pkce()}
            auth_url = auth_url % self.adding_extra['code_verifier']

        self.adding_allow = True
        webbrowser.open(auth_url, 2, True)

    def _clear_new_account(self):
        self.accounts_combo.set_active_id(None)
        self.username_entry.set_text("")
        self.password_entry.set_text("")

    def _show_oauth_account(self):
        self.username_label.set_text("Name")
        self.password_label.set_text("PIN")
        self.password_entry.set_visibility(True)
        self.btn_pin_request.show()

    def _show_password_account(self):
        self.username_label.set_text("Username")
        self.password_label.set_text("Password")
        self.password_entry.set_visibility(False)
        self.btn_pin_request.hide()

    def _get_combo_active_api_name(self):
        apiiter = self.accounts_combo.get_active_iter()

        if not apiiter:
            return None

        return self.accounts_combo.get_model().get(apiiter, 0)[0]

    @Gtk.Template.Callback()
    def _on_username_entry_changed(self, entry):
        if self.current == AccountsView.NEW:
            self._refresh_btn_new_confirm()
        elif self.current == AccountsView.EDIT:
            self._refresh_btn_edit_confirm()

    @Gtk.Template.Callback()
    def _on_password_entry_changed(self, entry):
        if self.current == AccountsView.NEW:
            self._refresh_btn_new_confirm()
        elif self.current == AccountsView.EDIT:
            self._refresh_btn_edit_confirm()

    def _refresh_btn_new_confirm(self):
        sensitive = (
            self.adding_allow and
            self._get_combo_active_api_name() and
            self.username_entry.get_text().strip() and
            self.password_entry.get_text()
        )

        self.btn_new_confirm.set_sensitive(sensitive)

    def _refresh_btn_edit_confirm(self):
        sensitive = (
            self.password_entry.get_text() and
            not self.account_edit['password'] == self.password_entry.get_text()
        )
        self.btn_edit_confirm.set_sensitive(sensitive)

    def _add_account(self):
        username = self.username_entry.get_text().strip()
        password = self.password_entry.get_text()
        api = self._get_combo_active_api_name()

        self.manager.add_account(username, password, api, self.adding_extra)

    def _edit_account(self):
        num = self.account_edit['account_id']
        username = self.account_edit['username']
        password = self.password_entry.get_text()
        api = self.account_edit['api']

        self.manager.edit_account(num, username, password, api)


@Gtk.Template.from_file(os.path.join(gtk_dir, 'data/accountrow.ui'))
class AccountRow(Gtk.ListBoxRow):
    __gtype_name__ = 'AccountRow'

    account_logo = Gtk.Template.Child()
    account_username = Gtk.Template.Child()
    account_api = Gtk.Template.Child()

    def __init__(self, account):
        Gtk.ListBoxRow.__init__(self)
        self.init_template()
        self.account = account

        if account['active']:
            self.account_username.set_text(account['username'])
            self.account_api.set_text(account['libname_desc'])
            self.account_logo.set_from_pixbuf(account['logo'])

    def get_account_id(self):
        return self.account['number']

    def get_libname(self):
        return self.account['libname']



import pickle

from trackma import utils


class AccountManager:
    """
    This is the account manager.

    It provides a generic way for the user interface to query for the
    available registered accounts, and add or delete accounts.

    This class returns an Account Dictionary used by
    the :class:`Engine` to start.
    """
    accounts = {'default': None, 'next': 1, 'accounts': dict()}

    def __init__(self):
        utils.make_dir(utils.to_config_path())
        self.filename = utils.to_config_path('accounts.dict')
        self._load()

    def _load(self):
        if utils.file_exists(self.filename):
            with open(self.filename, 'rb') as f:
                self.accounts = pickle.load(f)

    def _save(self):
        is_new = not utils.file_exists(self.filename)
        with open(self.filename, 'wb') as f:
            if is_new:
                utils.change_permissions(self.filename, 0o600)
            pickle.dump(self.accounts, f, protocol=2)

    def add_account(self, username, password, api, extra={}):
        """
        Registers a new account with the specified
        *username*, *password*, and *api*.

        The *api* must be one of the available APIs
        found in the utils.available_libs dict.
        """

        available_libs = utils.available_libs.keys()

        if not username:
            raise utils.AccountError('Empty username.')
        if not password:
            raise utils.AccountError('Empty password.')
        if api not in available_libs:
            raise utils.AccountError('That API doesn\'t exist.')

        account = {'username': username,
                   'password': password,
                   'api': api,
                   'extra': extra,
                   }

        nextnum = self.accounts['next']
        self.accounts['accounts'][nextnum] = account
        self.accounts['next'] += 1
        self._save()

    def edit_account(self, num, username, password, api, extra={}):
        """
        Updates data for account *num* with the specified
        *username*, *password*, and *api*.
        """

        available_libs = utils.available_libs.keys()

        if not username:
            raise utils.AccountError('Empty username.')
        if not password:
            raise utils.AccountError('Empty password.')
        if api not in available_libs:
            raise utils.AccountError('That API doesn\'t exist.')

        account = {'username': username,
                   'password': password,
                   'api': api,
                   'extra': extra,
                   }

        self.accounts['accounts'][num].update(account)
        self._save()

    def delete_account(self, num):
        """
        Deletes the account number **num**.
        """
        self.accounts['default'] = None
        del self.accounts['accounts'][num]

        # Reset index if there are no accounts left
        if not self.accounts['accounts']:
            self.accounts['next'] = 1

        self._save()

    def purge_account(self, num):
        """
        Renames stale cache files for account number **num**.
        """
        account = self.accounts['accounts'][num]
        userfolder = utils.to_data_path(
            "%s.%s" % (account['username'], account['api']))
        utils.make_dir(userfolder + '.old')
        utils.regex_rename_files(
            '(.*.queue)|(.*.info)|(.*.list)|(.*.meta)', userfolder, userfolder + '.old')

    def get_account(self, num):
        """
        Returns the account dict **num**.
        """
        return self.accounts['accounts'][num]

    def get_accounts(self):
        """
        Returns an iterator of available accounts.
        """
        return self.accounts['accounts'].items()

    def get_default(self):
        """
        Returns the default account number, if set.
        Otherwise returns None.
        """
        num = self.accounts['default']
        if num is not None:
            try:
                return self.accounts['accounts'][num]
            except KeyError:
                return None
        else:
            return None

    def set_default(self, val):
        """
        Sets a new default account number.
        """
        self.accounts['default'] = val
        self._save()

    def unset_default(self):
        """
        Unsets the default account number.
        """
        self.accounts['default'] = None



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import datetime
import json
import socket
import urllib.error
import urllib.parse
import urllib.request

from trackma import utils
from trackma.lib.lib import lib


class libanilist(lib):
    """
    API class to communicate with Anilist

    Website: https://anilist.co

    messenger: Messenger object to send useful messages to
    """
    name = 'libanilist'
    msg = None
    logged_in = False

    api_info = {'name': 'Anilist', 'shortname': 'anilist',
                'version': '2.1', 'merge': False}
    mediatypes = dict()
    mediatypes['anime'] = {
        'has_progress': True,
        'can_add': True,
        'can_delete': True,
        'can_score': True,
        'can_status': True,
        'can_update': True,
        'can_play': True,
        'can_date': True,
        'date_next_ep': True,
        'statuses_start': ['CURRENT', 'REPEATING'],
        'statuses_finish': ['COMPLETED'],
        'statuses_library': ['CURRENT', 'REPEATING', 'PAUSED', 'PLANNING'],
        'statuses':  ['CURRENT', 'COMPLETED', 'REPEATING', 'PAUSED', 'DROPPED', 'PLANNING'],
        'statuses_dict': {
            'CURRENT': 'Watching',
            'COMPLETED': 'Completed',
            'REPEATING': 'Rewatching',
            'PAUSED': 'Paused',
            'DROPPED': 'Dropped',
            'PLANNING': 'Plan to Watch'
        },
        'score_max': 100,
        'score_step': 1,
        'search_methods': [utils.SearchMethod.KW, utils.SearchMethod.SEASON],
    }
    mediatypes['manga'] = {
        'has_progress': True,
        'can_add': True,
        'can_delete': True,
        'can_score': True,
        'can_status': True,
        'can_update': True,
        'can_play': False,
        'can_date': True,
        'statuses_start': ['CURRENT', 'REPEATING'],
        'statuses_finish': ['COMPLETED'],
        'statuses':  ['CURRENT', 'COMPLETED', 'REPEATING', 'PAUSED', 'DROPPED', 'PLANNING'],
        'statuses_dict': {
            'CURRENT': 'Reading',
            'COMPLETED': 'Completed',
            'REPEATING': 'Rereading',
            'PAUSED': 'Paused',
            'DROPPED': 'Dropped',
            'PLANNING': 'Plan to Read'
        },
        'score_max': 100,
        'score_step': 1,
        'search_methods': [utils.SearchMethod.KW],
    }
    default_mediatype = 'anime'

    score_types = {
        'POINT_100': (100, 1),
        'POINT_10_DECIMAL': (10, 0.1),
        'POINT_10': (10, 1),
        'POINT_5': (5, 1),
        'POINT_3': (3, 1),
    }

    type_translate = {
        None: utils.Type.UNKNOWN,
        'TV': utils.Type.TV,
        'TV_SHORT': utils.Type.TV_SHORT,
        'MOVIE': utils.Type.MOVIE,
        'SPECIAL': utils.Type.SP,
        'OVA': utils.Type.OVA,
        'ONA': utils.Type.ONA,
        'MUSIC': utils.Type.OTHER,
        'MANGA': utils.Type.OTHER,
        'NOVEL': utils.Type.OTHER,
        'ONE_SHOT': utils.Type.OTHER,
    }

    status_translate = {
        None: utils.Status.UNKNOWN,
        'RELEASING': utils.Status.AIRING,
        'FINISHED': utils.Status.FINISHED,
        'NOT_YET_RELEASED': utils.Status.NOTYET,
        'CANCELLED': utils.Status.CANCELLED,
    }

    season_translate = {
        utils.Season.WINTER: 'WINTER',
        utils.Season.SPRING: 'SPRING',
        utils.Season.SUMMER: 'SUMMER',
        utils.Season.FALL: 'FALL',
    }
    rev_season_translate = {v: k for k, v in season_translate.items()}

    # Supported signals for the data handler
    signals = {'show_info_changed': None, }

    auth_url = "https://anilist.co/api/v2/"
    query_url = "https://graphql.anilist.co"
    client_id = "537"
    _client_secret = "9Hl31gyz2q9xMhhJwLKRA8DAn0pXl9sOHFf6I1YO"
    user_agent = 'Trackma/{}'.format(utils.VERSION)

    def __init__(self, messenger, account, userconfig):
        """Initializes the API"""
        super(libanilist, self).__init__(messenger, account, userconfig)

        self.pin = account['password'].strip()
        self.userid = self._get_userconfig('userid')

        if self.mediatype == 'manga':
            self.total_str = "chapters"
            self.watched_str = "chapters_read"
        else:
            self.total_str = "episodes"
            self.watched_str = "episodes_watched"

        # If we already know the scoreFormat of the cached list, apply it now
        self.scoreformat = self._get_userconfig(
            'scoreformat_' + self.mediatype)
        if self.scoreformat:
            self._apply_scoreformat(self.scoreformat)

        self.opener = urllib.request.build_opener()
        self.opener.addheaders = [('User-agent', self.user_agent)]

    def _raw_request(self, method, url, get=None, post=None, jsonpost=None, auth=False):
        if get:
            url = "{}?{}".format(url, urllib.parse.urlencode(get))
        if post:
            post = urllib.parse.urlencode(post).encode('utf-8')
        if jsonpost:
            post = json.dumps(jsonpost, ensure_ascii=False).encode('utf-8')

        request = urllib.request.Request(url, post)
        request.get_method = lambda: method

        request.add_header('Content-Type', 'application/json')
        request.add_header('Accept', 'application/json')

        if auth:
            request.add_header('Authorization', 'Bearer {}'.format(
                self.pin,
            ))

        try:
            response = self.opener.open(request, timeout=10)
            return json.loads(response.read().decode('utf-8'))
        except urllib.error.HTTPError as e:
            if e.code == 400:
                raise utils.APIError("Invalid HTTP request: %s" % e.read())
            else:
                raise utils.APIError("HTTP error status: %s" % e.read())
        except urllib.error.URLError as e:
            raise utils.APIError("HTTP connection error: %s" % e.reason)
        except socket.timeout:
            raise utils.APIError("Connection timed out.")

    def _request(self, query, variables=None):
        if variables:
            data = {'query': query, 'variables': variables}
        else:
            data = {'query': query}

        return self._raw_request('POST', self.query_url, jsonpost=data, auth=True)

    def check_credentials(self):
        if len(self.pin) == 40:  # Old pins were 40 digits, new ones seem to be 654 digits
            raise utils.APIFatal("This appears to be a V1 API PIN. You need a V2 API PIN to continue using AniList."
                                 " Please re-authorize or re-create your AniList account.")

        if not self.userid:
            self._refresh_user_info()

        return True

    def _refresh_user_info(self):
        self.msg.info('Refreshing user details...')
        query = '{Viewer{ id name avatar{large} options{titleLanguage displayAdultContent} mediaListOptions{scoreFormat} }}'
        data = self._request(query)['data']['Viewer']

        self._set_userconfig('userid', data['id'])
        self._set_userconfig('username', data['name'])
        self._emit_signal('userconfig_changed')

        self.userid = data['id']

    def fetch_list(self):
        self.check_credentials()
        self.msg.info('Downloading list...')

        query = '''query ($id: Int!, $listType: MediaType) {
  MediaListCollection (userId: $id, type: $listType) {
    lists {
      name
      isCustomList
      status
      entries {
        ... mediaListEntry
      }
    }
    user {
      mediaListOptions {
        scoreFormat
      }
    }
  }
}

fragment mediaListEntry on MediaList {
  id
  score
  progress
  startedAt { year month day }
  updatedAt
  completedAt { year month day }
  media {
    id
    title { userPreferred romaji english native }
    synonyms
    coverImage { large medium }
    format
    status
    chapters episodes
    nextAiringEpisode { airingAt episode }
    startDate { year month day }
    endDate { year month day }
    siteUrl
  }
}'''
        variables = {'id': self.userid, 'listType': self.mediatype.upper()}
        data = self._request(query, variables)['data']['MediaListCollection']

        showlist = {}

        if not data['lists']:
            # No lists returned so no need to continue
            return showlist

        # Handle different score formats provided by Anilist
        self.scoreformat = data['user']['mediaListOptions']['scoreFormat']
        self._apply_scoreformat(self.scoreformat)

        self._set_userconfig('scoreformat_' + self.mediatype, self.scoreformat)
        self._emit_signal('userconfig_changed')

        for remotelist in data['lists']:
            my_status = remotelist['status']

            if my_status not in self.media_info()['statuses']:
                continue
            if remotelist['isCustomList']:
                continue  # Maybe do something with this later
            for item in remotelist['entries']:
                show = utils.show()
                media = item['media']
                showid = media['id']
                showdata = {
                    'my_id': item['id'],
                    'id': showid,
                    'title': media['title']['userPreferred'],
                    'aliases': self._get_aliases(media),
                    'type': self._translate_type(media['format']),
                    'status': self._translate_status(media['status']),
                    'my_progress': self._c(item['progress']),
                    'my_status': my_status,
                    'my_score': self._c(item['score']),
                    'total': self._c(media[self.total_str]),
                    'image': media['coverImage']['large'],
                    'image_thumb': media['coverImage']['medium'],
                    'url': media['siteUrl'],
                    'start_date': self._dict2date(media['startDate']),
                    'end_date': self._dict2date(media['endDate']),
                    'my_start_date': self._dict2date(item['startedAt']),
                    'my_finish_date': self._dict2date(item['completedAt']),
                    'my_last_update': self._int2datetime(item['updatedAt']),
                }
                if media['nextAiringEpisode']:
                    showdata['next_ep_number'] = media['nextAiringEpisode']['episode']
                    showdata['next_ep_time'] = self._int2date(
                        media['nextAiringEpisode']['airingAt'])
                show.update({k: v for k, v in showdata.items() if v})
                showlist[showid] = show
        return showlist

    args_SaveMediaListEntry = {
        'id': 'Int',                         # The list entry id, required for updating
        'mediaId': 'Int',                    # The id of the media the entry is of
        'status': 'MediaListStatus',         # The watching/reading status
        'scoreRaw': 'Int',                   # The score of the media in 100 point
        # The amount of episodes/chapters consumed by the user
        'progress': 'Int',
        'startedAt': 'FuzzyDateInput',       # When the entry was started by the user
        'completedAt': 'FuzzyDateInput',     # When the entry was completed by the user
    }

    def _update_entry(self, item):
        """
        New entries will lack a list entry id, while updates will include one.
        In the case of a new entry, we want to record the new id. In the case of an update, we want to record the new updated date.
        """
        values = {'mediaId': item['id']}
        if 'my_id' in item and item['my_id']:
            values['id'] = item['my_id']
        if 'my_progress' in item:
            values['progress'] = item['my_progress']
        if 'my_status' in item:
            values['status'] = item['my_status']
        if 'my_score' in item:
            values['scoreRaw'] = self._score2raw(item['my_score'])
        if 'my_start_date' in item:
            values['startedAt'] = self._date2dict(item['my_start_date'])
        if 'my_finish_date' in item:
            values['completedAt'] = self._date2dict(item['my_finish_date'])

        vars_defn = ', '.join(
            ['${}: {}'.format(k, self.args_SaveMediaListEntry[k]) for k in values.keys()])
        subs_defn = ', '.join(['{0}: ${0}'.format(k) for k in values.keys()])
        query = 'mutation ({0}) {{ SaveMediaListEntry({1}) {{id updatedAt}} }}'.format(
            vars_defn, subs_defn)

        data = self._request(query, values)['data']
        return data['SaveMediaListEntry']

    def add_show(self, item):
        self.check_credentials()
        self.msg.info("Adding item %s..." % item['title'])
        return self._update_entry(item)['id']

    def update_show(self, item):
        self.check_credentials()
        self.msg.info("Updating item %s..." % item['title'])
        return self._int2datetime(self._update_entry(item)['updatedAt'])

    def delete_show(self, item):
        self.check_credentials()
        self.msg.info("Deleting item %s..." % item['title'])
        query = 'mutation ($id: Int) {DeleteMediaListEntry(id: $id){deleted} }'
        variables = {'id': item['my_id']}
        self._request(query, variables)

    def search(self, criteria, method):
        self.check_credentials()
        self.msg.info("Searching for {}...".format(criteria))

        if method == utils.SearchMethod.KW:
            query = "query ($query: String, $type: MediaType) { Page { media(search: $query, type: $type) {"
            variables = {'query': urllib.parse.quote_plus(criteria)}
        elif method == utils.SearchMethod.SEASON:
            season, seasonYear = criteria

            query = "query ($season: MediaSeason, $seasonYear: Int, $type: MediaType) { Page { media(season: $season, seasonYear: $seasonYear, type: $type) {"
            variables = {
                'season': self.season_translate[season], 'seasonYear': seasonYear}

        query += '''
      id
      title { userPreferred romaji english native }
      coverImage { medium large }
      format
      averageScore
      meanScore
      chapters episodes
      status
      startDate { year month day }
      endDate { year month day }
      siteUrl
      description
      genres
      synonyms
      studios(sort: NAME, isMain: true) { nodes { name } }
      seasonYear
      season
    }
  }
}'''
        variables['type'] = self.mediatype.upper()
        data = self._request(query, variables)['data']['Page']['media']

        infolist = []
        for media in data:
            infolist.append(self._parse_info(media))

        self._emit_signal('show_info_changed', infolist)
        return infolist

    def request_info(self, itemlist):
        self.check_credentials()
        infolist = []

        query = '''query ($id: Int!, $type: MediaType) {
  Media(id: $id, type: $type) {
      id
      title { userPreferred romaji english native }
      coverImage { medium large }
      format
      averageScore
      meanScore
      chapters episodes
      status
      startDate { year month day }
      endDate { year month day }
      siteUrl
      description
      genres
      synonyms
      studios(sort: NAME, isMain: true) { nodes { name } }
      seasonYear
      season
  }
}'''

        for show in itemlist:
            variables = {'id': show['id'], 'listType': self.mediatype.upper()}
            data = self._request(query, variables)['data']['Media']
            infolist.append(self._parse_info(data))

        self._emit_signal('show_info_changed', infolist)
        return infolist

    def media_info(self):
        """Return information about the currently selected mediatype."""
        return self.mediatypes[self.mediatype]

    def _parse_info(self, item):
        info = utils.show()
        showid = item['id']
        type_ = self._translate_type(item['format'])
        status = self._translate_status(item['status'])
        season = self.rev_season_translate.get(item.get('season'))
        if season and 'seasonYear' in item:
            season_and_year = f"{season!s} {item['seasonYear']}"
        elif 'seasonYear' in item:
            season_and_year = item['seasonYear']
        else:
            season_and_year = None

        info.update({
            'id': showid,
            'title': item['title']['userPreferred'],
            'total': self._c(item[self.total_str]),
            'aliases': self._get_aliases(item),
            'type': type_,
            'status': status,
            'image': item['coverImage']['large'],
            'image_thumb': item['coverImage']['medium'],
            'url': item['siteUrl'],
            'start_date': self._dict2date(item.get('startDate')),
            'end_date': self._dict2date(item.get('endDate')),
            'extra': [
                ('English',         item['title'].get('english')),
                ('Romaji',          item['title'].get('romaji')),
                ('Japanese',        item['title'].get('native')),
                ('Synonyms',        item.get('synonyms')),
                ('Season',          season_and_year),
                ('Genres',          item.get('genres')),
                ('Studios',         [s['name'] for s in item['studios']['nodes']]),
                ('Synopsis',        item.get('description')),
                ('Type',            type_),
                ('Average score',   item.get('averageScore')),
                ('Mean score',      item.get('meanScore')),
                ('Status',          status),
            ]
        })
        return info

    def _apply_scoreformat(self, fmt):
        media = self.media_info()
        (media['score_max'], media['score_step']) = self.score_types[fmt]

    def _get_aliases(self, item):
        aliases = [a for a in (item['title']['romaji'], item['title']
                               ['english'], item['title']['native']) if a] + item['synonyms']

        return aliases

    def _translate_type(self, orig_type):
        return self.type_translate.get(orig_type, utils.Type.UNKNOWN)

    def _translate_status(self, orig_status):
        return self.status_translate.get(orig_status, utils.Status.UNKNOWN)

    def _dict2date(self, item):
        if not item:
            return None
        try:
            return datetime.datetime(item['year'], item['month'], item['day'])
        except (TypeError, ValueError):
            return None

    def _date2dict(self, date):
        if not date:
            return {}
        try:
            return {'year': date.year, 'month': date.month, 'day': date.day}
        except (TypeError, ValueError):
            return {}

    def _score2raw(self, score):
        if score == 0:
            return 0

        if self.scoreformat in ['POINT_10', 'POINT_10_DECIMAL']:
            return int(score*10)
        elif self.scoreformat == 'POINT_5':
            return int(score*20)
        elif self.scoreformat == 'POINT_3':
            return int(score*25)
        else:
            return score

    def _int2date(self, item):
        if not item:
            return None
        try:
            return datetime.datetime.utcfromtimestamp(item)
        except ValueError:
            return None

    def _int2datetime(self, item):
        if not item:
            return None
        try:
            return datetime.datetime.fromtimestamp(item, tz=datetime.timezone.utc)
        except ValueError:
            return None

    def _c(self, s):
        if s is None:
            return 0
        else:
            return s



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


import os
import sys

from trackma import utils
from trackma.ui.qt.mainwindow import MainWindow


def main():
    print("Trackma-qt v{}".format(utils.VERSION))

    debug = False

    if '-h' in sys.argv:
        print("Usage: trackma-qt [options]")
        print()
        print('Options:')
        print(' -d  Shows debugging information')
        print(' -h  Shows this help')
        sys.exit(0)
    if '-d' in sys.argv:
        print('Showing debug information.')
        debug = True

    try:
        from PyQt6.QtWidgets import QApplication, QMessageBox
    except ImportError:
        print("Couldn't import Qt6 dependencies. "
              "Make sure you installed the PyQt6 package.")

    try:
        from PIL import Image
        os.environ['imaging_available'] = "1"
    except ImportError:
        print("Warning: PIL or Pillow isn't available. "
              "Preview images will be disabled.")

    app = QApplication(sys.argv)
    app.setApplicationName("trackma")
    app.setDesktopFileName("trackma-qt")
    if os.name == "nt":
        import ctypes
        myappid = 'trackma' + utils.VERSION
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
    try:
        # keep the variable around to prevent it from being gc'ed
        main_window = MainWindow(debug)
        sys.exit(app.exec())
    except utils.TrackmaFatal as e:
        QMessageBox.critical(None, 'Fatal Error', "{0}".format(e), QMessageBox.StandardButton.Ok)



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import copy
import datetime
import difflib
import json
import locale
import os
import pickle
import re
import shutil
import subprocess
import sys
import threading
import time
import uuid
from enum import Enum, auto

VERSION = '0.10.1'

DATADIR = os.path.dirname(__file__) + '/data'


class Login(Enum):
    PASSWD = auto()
    OAUTH = auto()
    OAUTH_PKCE = auto()


class BaseEnum(Enum):
    @classmethod
    def find(cls, name):
        try:
            return cls(name)
        except ValueError:
            pass
        try:
            return cls[name.upper().replace(' ', '_')]
        except KeyError:
            pass
        return cls.UNKNOWN

    @classmethod
    def from_int(cls, index):
        try:
            return list(cls.__members__.values())[index]
        except IndexError:
            return cls.UNKNOWN

    def __int__(self):
        try:
            return list(self.__class__.__members__.values()).index(self)
        except IndexError:
            return list(self.__class__.__members__.values()).index(self.__class__.UNKNOWN)

    def __lt__(self, other):
        return int(self) < int(other)

    def __le__(self, other):
        return int(self) <= int(other)

    def __gt__(self, other):
        return int(self) > int(other)

    def __ge__(self, other):
        return int(self) >= int(other)

    def __add__(self, other):
        if isinstance(other, str):
            return str(self) + other
        else:
            return super().__add__(other)

    def __str__(self):
        if isinstance(self.value, (str,)):
            return self.value
        else:
            return self.name.replace('_', ' ')


class Status(BaseEnum):
    UNKNOWN = 'Unknown'
    ONGOING = 'Ongoing'
    FINISHED = 'Finished'
    NOTYET = 'Not yet started'
    CANCELLED = 'Cancelled'
    OTHER = 'Other'

    # aliases
    AIRING = ONGOING
    RELEASING = ONGOING
    PUBLISHING = ONGOING
    CURRENTLY_AIRING = ONGOING
    FINISHED_AIRING = FINISHED
    NOT_YET_AIRED = NOTYET
    NOT_YET_RELEASED = NOTYET
    NOT_YET_PUBLISHED = NOTYET


class Type(BaseEnum):
    UNKNOWN = "Unknown"
    OTHER = "Other"

    # anime
    TV = "TV"
    TV_SHORT = "TV Short"
    MOVIE = "Movie"
    OVA = "OVA"
    ONA = "ONA"
    SPECIAL = "Special"

    # manga
    MANGA = "Manga"
    NOVEL = "Novel"
    ONE_SHOT = "One Shot"

    # aliases
    SP = SPECIAL
    MUSIC = OTHER


class Tracker(Enum):
    NOVIDEO = auto()
    PLAYING = auto()
    UNRECOGNIZED = auto()
    NOT_FOUND = auto()
    IGNORED = auto()


class Season(BaseEnum):
    WINTER = 'Winter'
    SPRING = 'Spring'
    SUMMER = 'Summer'
    FALL = 'Fall'


class SearchMethod(Enum):
    KEYWORD = auto()
    SEASON = auto()

    # aliases
    KW = KEYWORD


HOME = os.path.expanduser("~")
EXTENSIONS = ('.mkv', '.mp4', '.avi', '.ts')

# Put the available APIs here
available_libs = {
    'anilist':   ('Anilist',      DATADIR + '/anilist.jpg',     Login.OAUTH,
                  "https://anilist.co/api/v2/oauth/authorize?client_id=537&response_type=token"),
    'kitsu':     ('Kitsu',        DATADIR + '/kitsu.png',       Login.PASSWD),
    'mal':       ('MyAnimeList',  DATADIR + '/mal.jpg',     Login.OAUTH_PKCE,
                  "https://myanimelist.net/v1/oauth2/authorize?response_type=code&client_id=32c510ab2f47a1048a8dd24de266dc0c&code_challenge=%s"),
    'shikimori': ('Shikimori',    DATADIR + '/shikimori.jpg',   Login.OAUTH,
                  "https://shikimori.one/oauth/authorize?client_id=Jfu9MKkUKPG4fOC95A6uwUVLHy3pwMo3jJB7YLSp7Ro&redirect_uri=urn%3Aietf%3Awg%3Aoauth%3A2.0%3Aoob&response_type=code&scope=user_rates"),
    'vndb':      ('VNDB',         DATADIR + '/vndb.jpg',        Login.PASSWD),
}

available_trackers = [
    ('auto', 'Auto-detect'),
    ('inotify_auto', 'inotify'),
    ('polling', 'Polling (lsof)'),
    ('mpris', 'MPRIS'),
    ('plex', 'Plex Media Server'),
    ('jellyfin', 'Jellyfin'),
    ('kodi', 'Kodi'),
    ('win32', 'Win32'),
]

available_parsers = [
    ('aie', 'AnimeInfoExtractor (Default)'),
    ('anitopy', 'Anitopy (Experimental)'),
]

def oauth_generate_pkce() -> str:
    import secrets

    token = secrets.token_urlsafe(100)
    return token[:128]


def parse_config(filename, default):
    config = copy.copy(default)

    try:
        with open(filename) as configfile:
            loaded_config = json.load(configfile)
            if 'colors' in config and 'colors' in loaded_config:
                # Need to prevent nested dict from being overwritten with an incomplete dict
                config['colors'].update(loaded_config['colors'])
                loaded_config['colors'] = config['colors']
            config.update(loaded_config)
    except IOError:
        # Will just use the default config
        # and create the file for manual editing
        save_config(config, filename)
    except ValueError:
        # There's a syntax error in the config file
        errorString = "Erroneous config %s requires manual fixing or deletion to proceed." % filename
        log_error(errorString)
        raise TrackmaFatal(errorString)

    return config


def save_config(config_dict, filename):
    path = os.path.dirname(filename)
    if not os.path.isdir(path):
        os.mkdir(path)

    with open(filename, 'wb') as configfile:
        configfile.write(json.dumps(config_dict, sort_keys=True,
                                    indent=4, separators=(',', ': ')).encode('utf-8'))


def load_data(filename):
    with open(filename, 'rb') as datafile:
        return pickle.load(datafile, encoding='bytes')


def save_data(data, filename):
    with open(filename, 'wb') as datafile:
        pickle.dump(data, datafile, protocol=2)


def log_error(msg):
    with open(to_data_path('error.log'), 'a') as logfile:
        logfile.write(msg)


def expand_path(path):
    return os.path.expanduser(path)


def expand_paths(paths):
    return (expand_path(path) for path in paths)


def is_media(filename):
    return os.path.splitext(filename)[1] in EXTENSIONS


def get_any(dict_, *keys, default=None):
    """Get the first key present on a dict."""
    for key in keys:
        value = dict_.get(key, Ellipsis)
        if value is not Ellipsis:
            return value
    return default


def regex_find_videos(subdirectory=''):
    if subdirectory:
        path = os.path.expanduser(subdirectory)
    else:
        path = os.getcwd()

    for root, dirs, names in os.walk(path, followlinks=True):
        for filename in names:
            if is_media(filename):
                yield (os.path.join(root, filename), filename)


def regex_rename_files(pattern, source_dir, dest_dir):
    in_path = os.path.expanduser(os.path.join('~', '.trackma', source_dir))
    out_path = os.path.expanduser(os.path.join('~', '.trackma', dest_dir))
    for filename in os.listdir(in_path):
        if re.match(pattern, filename):
            in_file = os.path.join(in_path, filename)
            out_file = os.path.join(out_path, filename)
            os.rename(in_file, out_file)


def list_library(path):
    for root, dirs, names in os.walk(path, followlinks=True):
        for filename in names:
            yield (os.path.join(root, filename), filename)


def make_dir(path):
    if not os.path.isdir(path):
        os.makedirs(path)


def dir_exists(dirname):
    return os.path.isdir(dirname)


def file_exists(filename):
    return os.path.isfile(filename)


def file_older_than(filename, mtime):
    return os.path.getmtime(filename) < time.time() - mtime


def try_files(filenames):
    for filename in filenames:
        if file_exists(filename):
            return filename


def sync_file(fname, sync_url):
    if not sync_url:
        return False

    import urllib.request
    import socket

    try:
        with urllib.request.urlopen(sync_url) as r, open(fname, 'wb') as f:
            shutil.copyfileobj(r, f)
    except (socket.timeout, urllib.error.URLError):
        return False

    return True


def copy_file(src, dest):
    shutil.copy(src, dest)


def to_config_path(*paths):
    if dir_exists(os.path.join(HOME, ".trackma")):
        return os.path.join(HOME, ".trackma", *paths)

    return os.path.join(os.environ.get("XDG_CONFIG_HOME", os.path.join(HOME, ".config")), "trackma", *paths)


def to_data_path(*paths):
    if dir_exists(os.path.join(HOME, ".trackma")):
        return os.path.join(HOME, ".trackma", *paths)

    return os.path.join(os.environ.get("XDG_DATA_HOME", os.path.join(HOME, ".local", "share")), "trackma", *paths)


def to_cache_path(*paths):
    if dir_exists(os.path.join(HOME, ".trackma")):
        return os.path.join(HOME, ".trackma", "cache", *paths)

    return os.path.join(os.environ.get("XDG_CACHE_HOME", os.path.join(HOME, ".cache")), "trackma", *paths)


def change_permissions(filename, mode):
    os.chmod(filename, mode)


def estimate_aired_episodes(show):
    """ Estimate how many episodes have passed since airing """

    if show['status'] == Status.FINISHED:
        return show['total']
    elif show['status'] == Status.NOTYET:
        return 0
    # It's airing, so we make an estimate based on available information
    elif show['status'] == Status.AIRING:
        if 'next_ep_number' in show:  # Do we have the upcoming episode number?
            return show['next_ep_number'] - 1
        elif show['start_date']:  # Do we know when it started? Let's just assume 1 episode = 1 week
            days = (datetime.datetime.now() - show['start_date']).days
            if days <= 0:
                return 0

            eps = days // 7 + 1
            if show['total'] and eps > show['total']:
                return show['total']
            return eps
    return 0


def guess_show(show_title, tracker_list):
    """ Take a title and search for it fuzzily in the tracker list """
    (showlist, altnames_map) = tracker_list

    # Return the show immediately if we find an altname for it
    if altnames_map and show_title.lower() in altnames_map:
        showid = altnames_map[show_title.lower()]

        if showid in showlist:
            return showlist[showid]

    # Use difflib to see if the show title is similar to
    # one we have in the list
    highest_ratio = (None, 0)
    matcher = difflib.SequenceMatcher()
    matcher.set_seq1(show_title.lower())

    # Compare to every show in our list to see which one
    # has the most similar name
    for item in showlist.values():
        # Make sure to search through all the aliases
        for title in item['titles']:
            matcher.set_seq2(title.lower())
            ratio = matcher.ratio()
            if ratio > highest_ratio[1]:
                highest_ratio = (item, ratio)

    playing_show = highest_ratio[0]
    if highest_ratio[1] > 0.7:
        return playing_show


def redirect_show(show_tuple, redirections, tracker_list):
    """ Use a redirection dictionary and return the new show ID and episode accordingly """
    if not redirections:
        return show_tuple

    (show, ep) = show_tuple
    if show.get('total') and ep <= show['total']:
        # redirect only for invalid ep
        return show_tuple

    showlist = tracker_list[0]
    if show['id'] in redirections:
        for redirection in redirections[show['id']]:
            (src_eps, dst_id, dst_eps) = redirection

            if (src_eps[1] == -1 and ep >= src_eps[0]) or (ep in range(src_eps[0], src_eps[1] + 1)):
                new_show_id = dst_id
                new_ep = ep + (dst_eps[0] - src_eps[0])

                if new_show_id in showlist:
                    return (showlist[new_show_id], new_ep)

    return show_tuple


def open_folder(path):
    if sys.platform == 'darwin':
        spawn_process(["open", path])
    elif sys.platform == 'win32':
        spawn_process(["explorer", path])
    else:
        spawn_process(["xdg-open", path])


_spawned_processes = []


def _poll_spawned_processes():
    """Periodically poll and prune process handles to prevent zombies."""
    while True:
        time.sleep(5)
        _spawned_processes[:] = [child for child in _spawned_processes if child.poll() is None]


_process_collector_thread = None


def spawn_process(arg_list):
    """
    Helper generic function to spawn a subprocess.

    The handles are collected and periodically polled
    to prevent zombie processes on *NIX.
    """
    global _process_collector_thread

    if not arg_list:
        return

    proc = subprocess.Popen(arg_list, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    _spawned_processes.append(proc)

    if not _process_collector_thread:
        _process_collector_thread = threading.Thread(target=_poll_spawned_processes, daemon=True)
        _process_collector_thread.start()


def get_terminal_size(fd=1):
    """
    Returns height and width of current terminal. First tries to get
    size via termios.TIOCGWINSZ, then from environment. Defaults to 25
    lines x 80 columns if both methods fail.

    :param fd: file descriptor (default: 1=stdout)
    """
    try:
        import fcntl
        import termios
        import struct
        hw = struct.unpack('hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
    except Exception:
        try:
            width = int(os.environ.get('COLUMNS', 80))
            height = int(os.environ.get('LINES', 25))
            hw = (height, width)
        except Exception:
            hw = (25, 80)

    return hw


def show():
    return {
        'id':           0,
        'title':        '',
        'url':          '',
        'aliases':      [],
        'my_id':        None,
        'my_progress':  0,
        'my_status':    1,
        'my_score':     0,
        'my_start_date':  None,
        'my_finish_date': None,
        'type':         0,
        'status':       0,
        'total':        0,
        'start_date':   None,
        'end_date':     None,
        'image':        '',
        'image_thumb':  '',
        'queued':       False,
        'my_last-update': None
    }


class TrackmaError(Exception):
    pass


class EngineError(TrackmaError):
    pass


class DataError(TrackmaError):
    pass


class APIError(TrackmaError):
    pass


class AccountError(TrackmaError):
    pass


class TrackmaFatal(Exception):
    pass


class EngineFatal(TrackmaFatal):
    pass


class DataFatal(TrackmaFatal):
    pass


class APIFatal(TrackmaFatal):
    pass


# Configuration defaults
config_defaults = {
    'player': 'mpv',
    'searchdir': ['~/Videos'],
    'tracker_enabled': True,
    'tracker_update_wait_s': 120,
    'tracker_update_close': False,
    'tracker_update_prompt': False,
    'tracker_not_found_prompt': False,
    'tracker_interval': 10,
    'tracker_process': 'mplayer|mplayer2|mpv',
    'tracker_ignore_not_next': True,
    'autoretrieve': 'days',
    'autoretrieve_days': 3,
    'autosend': 'minutes',
    'autosend_minutes': 60,
    'autosend_size': 5,
    'autosend_at_exit': True,
    'library_autoscan': True,
    'library_full_path': False,
    'scan_whole_list': False,
    'debug_disable_lock': True,
    'auto_status_change': True,
    'auto_status_change_if_scored': True,
    'auto_date_change': True,
    'tracker_type': "auto",
    'plex_host': "localhost",
    'plex_port': "32400",
    'plex_obey_update_wait_s': False,
    'plex_user': '',
    'plex_passwd': '',
    'plex_uuid': str(uuid.uuid1()),
    'plex_ssl': False,
    'jellyfin_host': "localhost",
    'jellyfin_port': "8096",
    'jellyfin_api_key': '',
    'jellyfin_user': '',
    'kodi_host': "localhost",
    'kodi_port': "8080",
    'kodi_obey_update_wait_s': False,
    'kodi_user': '',
    'kodi_passwd': '',
    'redirections_url': 'https://raw.githubusercontent.com/erengy/anime-relations/master/anime-relations.txt',
    'redirections_time': 1,
    'use_hooks': True,
    'title_parser': 'aie',
}

userconfig_defaults = {
    'mediatype': '',
    'userid': 0,
    'username': '',
}

curses_defaults = {
    'show_help': True,
    'keymap': {
        'help': '?',
        'prev_filter': 'left',
        'next_filter': 'right',
        'sort': 'f3',
        'sort_order': 'r',
        'update': 'u',
        'play': 'p',
        'openfolder': 'o',
        'play_random': '&',
        'status': 'f6',
        'score': 'z',
        'send': 's',
        'retrieve': 'R',
        'addsearch': 'a',
        'reload': 'c',
        'switch_account': 'f9',
        'delete': 'd',
        'quit': 'q',
        'altname': 'A',
        'search': '/',
        'neweps': 'N',
        'details': 'enter',
        'details_exit': 'esc',
        'open_web': 'O',
        'left': 'h',
        'up': 'k',
        'down': 'j',
        'right': 'l',
        'page_up': 'K',
        'page_down': 'J',
    },
    'palette': {
        'body':             ('', ''),
        'focus':            ('standout', ''),
        'head':             ('light red', 'black'),
        'header':           ('bold', ''),
        'status':           ('white', 'dark blue'),
        'error':            ('light red', 'dark blue'),
        'window':           ('white', 'dark blue'),
        'button':           ('black', 'light gray'),
        'button hilight':   ('white', 'dark red'),
        'item_airing':      ('dark blue', ''),
        'item_notaired':    ('yellow', ''),
        'item_neweps':      ('white', 'brown'),
        'item_updated':     ('white', 'dark green'),
        'item_playing':     ('white', 'dark blue'),
        'info_title':       ('light red', ''),
        'info_section':     ('dark blue', ''),
    }
}

gtk_defaults = {
    'show_tray': True,
    'close_to_tray': True,
    'start_in_tray': False,
    'tray_api_icon': False,
    'remember_geometry': False,
    'last_width': 740,
    'last_height': 480,
    'visible_columns': ['Title', 'Progress', 'Score', 'Percent'],
    'episodebar_style': 1,
    'colors': {
        'is_airing': '#0099CC',
        'is_playing': '#6C2DC7',
        'is_queued': '#54C571',
        'new_episode': '#FBB917',
        'not_aired': '#999900',
        'progress_bg': '#E5E5E5',
        'progress_fg': '#99B3CC',
        'progress_sub_bg': '#B3B3B3',
        'progress_sub_fg': '#668099',
        'progress_complete': '#99CCB3',
    },
}

qt_defaults = {
    'show_tray': True,
    'close_to_tray': True,
    'notifications': True,
    'start_in_tray': False,
    'tray_api_icon': False,
    'remember_geometry': False,
    'remember_columns': False,
    'last_x': 0,
    'last_y': 0,
    'last_width': 740,
    'last_height': 480,
    'visible_columns': ['Title', 'Progress', 'Score', 'Percent'],
    'inline_edit': True,
    'columns_state': None,
    'columns_per_api': False,
    'episodebar_style': 1,
    'episodebar_text': False,
    'filter_bar_position': 2,
    'filter_global': False,
    'colors': {
        'is_airing': '#D2FAFA',
        'is_playing': '#9696FA',
        'is_queued': '#D2FAD2',
        'new_episode': '#FAFA82',
        'not_aired': '#FAFAD2',
        'progress_bg': '#F5F5F5',
        'progress_fg': '#74C0FA',
        'progress_sub_bg': '#D2D2D2',
        'progress_sub_fg': '#5187B1',
        'progress_complete': '#00D200',
    },
    'sort_index': 1,
    'sort_order': 0,
}

qt_per_api_defaults = {
    'visible_columns': ['Title', 'Progress', 'Score', 'Percent'],
    'columns_state': None,
}

def format_local_time(
    dt: datetime.datetime | None,
    fallback_message='No data',
    error_message='?'
) -> str:
    if dt is None or dt.tzinfo is None:
        return fallback_message
    try:
        return dt.astimezone().strftime(locale.nl_langinfo(locale.D_T_FMT))
    except ValueError:
        return error_message



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import re
import time

from trackma import utils
from trackma.tracker import tracker


class inotifyBase(tracker.TrackerBase):
    open_file = (None, None, None)

    def __init__(self, messenger, tracker_list, config, watch_dirs, redirections=None):
        super().__init__(messenger, tracker_list, config, watch_dirs, redirections)

        self.re_players = re.compile(config['tracker_process'].encode('utf-8'))

    def _proc_poll(self):
        """
        This function scans proc to see if there's any player
        already open. If it is, and it has a media file open,
        return its first instance as a filename.
        """

        time.sleep(0.01)

        for p in os.listdir("/proc/"):
            if not p.isdigit():
                continue

            # Get process name
            try:
                with open('/proc/%s/cmdline' % p, 'rb') as f:
                    cmdline = f.read()
                    pname = cmdline.partition(b'\x00')[0]
            except FileNotFoundError:
                continue
            # It's not one of our players
            if not self.re_players.search(pname):
                continue

            d = "/proc/%s/fd/" % p
            try:
                for fd in os.listdir(d):
                    f = os.readlink(d+fd)
                    if utils.is_media(f):
                        return os.path.split(f)
            except OSError:
                pass

        return None

    def _is_being_played(self, filename):
        """
        This function makes sure that the filename is being played
        by the player specified in players.

        It uses procfs so if we're using inotify that means we're using Linux
        thus we should be safe.
        """

        time.sleep(0.01)
        for p in os.listdir("/proc/"):
            if not p.isdigit():
                continue
            d = "/proc/%s/fd/" % p
            try:
                for fd in os.listdir(d):
                    f = os.readlink(d+fd)
                    if f == filename:
                        # Get process name
                        with open('/proc/%s/cmdline' % p, 'rb') as f:
                            cmdline = f.read()
                            pname = cmdline.partition(b'\x00')[0]
                        self.msg.debug('Playing process: {} {} ({})'.format(p, pname, cmdline))

                        # Check if it's our process
                        if self.re_players.search(pname):
                            return p, fd
                        else:
                            self.msg.debug("Not read by player ({})".format(pname))
            except OSError:
                pass

        self.msg.debug("Couldn't find playing process.")
        return None, None

    def _closed_handle(self, pid, fd):
        """ Check if this pid has closed this handle (or never opened it) """
        d = "/proc/%s/fd/%s" % (pid, fd)
        # TODO : If we don't wait the filehandle will still be there
        time.sleep(0.01)
        return not os.path.islink(d)

    def _proc_open(self, path, name):
        self.msg.debug('Got OPEN event: {} {}'.format(path, name))
        pathname = os.path.join(path, name)

        pid, fd = self._is_being_played(pathname)

        if pid:
            self._emit_signal('detected', path, name)
            self.open_file = (pathname, pid, fd)

            if self.config['library_full_path']:
                (state, show_tuple) = self._get_playing_show(pathname)
            else:
                (state, show_tuple) = self._get_playing_show(name)
            self.msg.debug("Got status: {} {}".format(state, show_tuple))
            self.update_show_if_needed(state, show_tuple)
        else:
            self.msg.debug("Not played by player, ignoring.")

    def _proc_close(self, path, name):
        self.msg.debug('Got CLOSE event: {} {}'.format(path, name))
        pathname = os.path.join(path, name)

        open_pathname, pid, fd = self.open_file

        if pathname != open_pathname:
            self.msg.debug("A different file was closed.")
            return

        if not self._closed_handle(pid, fd):
            self.msg.debug("Our pid hasn't closed the file.")
            return

        self._emit_signal('detected', path, name)
        self.open_file = (None, None, None)

        (state, show_tuple) = self._get_playing_show(None)
        self.update_show_if_needed(state, show_tuple)



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import threading
import urllib.request
from io import BytesIO

from gi.repository import GLib, GdkPixbuf, Gtk

from trackma import utils

try:
    from PIL import Image
    imaging_available = True
except ImportError:
    imaging_available = False


class ImageThread(threading.Thread):
    def __init__(self, url, filename, width, height, callback):
        threading.Thread.__init__(self)
        self._url = url
        self._filename = filename
        self._width = width
        self._height = height
        self._callback = callback
        self._stop_request = threading.Event()

    def run(self):
        self._save_image(self._download_file())

        if self._stop_request.is_set():
            return

        if os.path.exists(self._filename):
            GLib.idle_add(self._callback, self._filename)

    def _download_file(self):
        request = urllib.request.Request(self._url)
        request.add_header(
            "User-Agent", "TrackmaImage/{}".format(utils.VERSION))
        return BytesIO(urllib.request.urlopen(request).read())

    def _save_image(self, img_bytes):
        if imaging_available:
            image = Image.open(img_bytes)
            image.thumbnail((self._width, self._height), Image.BICUBIC)
            image.convert("RGB").save(self._filename)
        else:
            with open(self._filename, 'wb') as img_file:
                img_file.write(img_bytes.read())

    def stop(self):
        self._stop_request.set()


class ImageBox(Gtk.HBox):
    def __init__(self, width, height):
        Gtk.HBox.__init__(self)

        self._width = width
        self._height = height

        self._image = Gtk.Image()
        self._image.set_size_request(width, height)

        self._label_holder = Gtk.Label()
        self._label_holder.set_size_request(width, height)

        self._image_thread = None

        if imaging_available:
            self.pack_start(self._label_holder, False, False, 0)
            self.pack_start(self._image, False, False, 0)
        else:
            self.pack_start(self._label_holder, False, False, 0)

        self.reset()

    def reset(self):
        if imaging_available:
            self.set_image(utils.DATADIR + '/icon.png')
        else:
            self.set_text("PIL library\nnot available")

    def set_text(self, text):
        self._label_holder.set_text(text)
        self._label_holder.show()
        self._image.hide()

    def set_image(self, filename):
        if not imaging_available:
            return

        pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename)
        width, height = scale(pixbuf.get_width(),
                              pixbuf.get_height(), self._width, self._height)
        scaled_buf = pixbuf.scale_simple(
            width, height, GdkPixbuf.InterpType.BILINEAR)

        self._image.set_from_pixbuf(scaled_buf)
        self._image.show()
        self._label_holder.hide()

    def set_image_remote(self, url, filename):
        if not imaging_available:
            return

        if self._image_thread:
            self._image_thread.stop()

        self.set_text("Loading...")
        self._image_thread = ImageThread(
            url, filename, self._width, self._height, self.set_image)
        self._image_thread.start()


def scale(w, h, x, y, maximum=True):
    nw = y * w / h
    nh = x * h / w
    if maximum ^ (nw >= x):
        return nw or 1, y
    return x, nh or 1



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import urllib.error
import urllib.request
from io import BytesIO

from PIL import Image
from PyQt6 import QtCore

from trackma import utils
from trackma.engine import Engine


class ImageWorker(QtCore.QThread):
    """
    Image thread

    Downloads an image and shrinks it if necessary.

    """
    cancelled = False
    finished = QtCore.pyqtSignal(str)

    def __init__(self, remote, local, size=None):
        self.remote = remote
        self.local = local
        self.size = size
        super(ImageWorker, self).__init__()

    def __del__(self):
        self.wait()

    def run(self):
        self.cancelled = False

        req = urllib.request.Request(self.remote)
        req.add_header("User-agent", "TrackmaImage/{}".format(utils.VERSION))
        try:
            img_file = BytesIO(urllib.request.urlopen(req).read())
            if self.size:
                if "imaging_available" in os.environ:
                    im = Image.open(img_file)
                    im.thumbnail((self.size[0], self.size[1]), Image.BICUBIC)
                    im.convert("RGB").save(self.local)
            else:
                with open(self.local, 'wb') as f:
                    f.write(img_file.read())
        except urllib.error.URLError as e:
            print("Warning: Error getting image ({})".format(e))
            return

        if self.cancelled:
            return

        self.finished.emit(self.local)

    def cancel(self):
        self.cancelled = True


class EngineWorker(QtCore.QThread):
    """
    Worker thread

    Contains the engine and manages every process in a separate thread.

    """
    engine = None
    function = None
    finished = QtCore.pyqtSignal(dict)

    # Message handler signals
    changed_status = QtCore.pyqtSignal(str, int, str)
    raised_error = QtCore.pyqtSignal(str)
    raised_fatal = QtCore.pyqtSignal(str)

    # Event handler signals
    changed_show = QtCore.pyqtSignal(dict)
    changed_show_status = QtCore.pyqtSignal(dict, object)
    changed_list = QtCore.pyqtSignal(dict)
    changed_queue = QtCore.pyqtSignal(int)
    tracker_state = QtCore.pyqtSignal(dict)
    playing_show = QtCore.pyqtSignal(dict, bool, int)
    prompt_for_update = QtCore.pyqtSignal(dict, int)
    prompt_for_add = QtCore.pyqtSignal(dict, int)

    def __init__(self):
        super(EngineWorker, self).__init__()

        self.overrides = {'start': self._start}

    def _messagehandler(self, classname, msgtype, msg):
        self.changed_status.emit(classname, msgtype, msg)

    def _error(self, msg):
        self.raised_error.emit(str(msg))

    def _fatal(self, msg):
        self.raised_fatal.emit(str(msg))

    def _changed_show(self, show, changes=None):
        self.changed_show.emit(show)

    def _changed_show_status(self, show, old_status=None):
        self.changed_show_status.emit(show, old_status)

    def _changed_list(self, show):
        self.changed_list.emit(show)

    def _changed_queue(self, queue):
        self.changed_queue.emit(len(queue))

    def _tracker_state(self, status):
        self.tracker_state.emit(status)

    def _playing_show(self, show, is_playing, episode):
        self.playing_show.emit(show, is_playing, episode)

    def _prompt_for_update(self, show, episode):
        self.prompt_for_update.emit(show, episode)

    def _prompt_for_add(self, show, episode):
        self.prompt_for_add.emit(show, episode)

    def _start(self, account):
        self.engine = Engine(account, self._messagehandler)

        self.engine.connect_signal('episode_changed', self._changed_show)
        self.engine.connect_signal('score_changed', self._changed_show)
        self.engine.connect_signal('tags_changed', self._changed_show)
        self.engine.connect_signal('status_changed', self._changed_show_status)
        self.engine.connect_signal('playing', self._playing_show)
        self.engine.connect_signal('show_added', self._changed_list)
        self.engine.connect_signal('show_deleted', self._changed_list)
        self.engine.connect_signal('show_synced', self._changed_show)
        self.engine.connect_signal('queue_changed', self._changed_queue)
        self.engine.connect_signal(
            'prompt_for_update', self._prompt_for_update)
        self.engine.connect_signal('prompt_for_add', self._prompt_for_add)
        self.engine.connect_signal('tracker_state', self._tracker_state)

        self.engine.start()

    def set_function(self, function, ret_function, *args, **kwargs):
        if function in self.overrides:
            self.function = self.overrides[function]
        else:
            self.function = getattr(self.engine, function)

        try:
            self.finished.disconnect()
        except Exception:
            pass

        if ret_function:
            self.finished.connect(ret_function)

        self.args = args
        self.kwargs = kwargs

    def __del__(self):
        self.wait()

    def run(self):
        try:
            ret = self.function(*self.args, **self.kwargs)
            self.finished.emit({'success': True, 'result': ret})
        except utils.TrackmaError as e:
            self._error(e)
            self.finished.emit({'success': False})
        except utils.TrackmaFatal as e:
            self._fatal(e)



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import base64
import json
import time
import urllib.error
import urllib.request

from trackma.tracker import tracker

NOT_RUNNING = 0
IDLE = 1
ACTIVE = 2
AUTH_REQUIRED = 3

PLAYING = 4
PAUSED = 5


class KodiTracker(tracker.TrackerBase):
    name = 'Tracker (Kodi)'

    def __init__(self, messenger, tracker_list, config, watch_dirs, redirections=None):
        self.config = config

        self.host_port = "{}:{}".format(
            self.config['kodi_host'], self.config['kodi_port'])
        self.status_log = [None, None]
        self.headers = {'content-type': 'application/json'}
        super().__init__(messenger, tracker_list, config, watch_dirs, redirections)

    def _get_kodi_status(self):
        # returns the status of the active players (if any)
        try:
            player = self._get_rpc_info("Player.GetActivePlayers")

            if player:
                if player[0]['type'] == "video":
                    return ACTIVE
                else:
                    return IDLE
            else:
                return IDLE
        except urllib.error.URLError as e:
            if hasattr(e, 'code'):
                if e.code == 401:
                    return AUTH_REQUIRED
                else:
                    return NOT_RUNNING
            else:
                return NOT_RUNNING

    def _playing_file(self):
        # returns the filename of the currently playing file
        if self._get_kodi_status() == IDLE:
            return None

        name = self._get_rpc_info("Player.GetItem", {"playerid": 1})[
            'item']['label']
        rstate = self._get_player_props("speed")

        if rstate > 0:
            state = PLAYING
        else:
            state = PAUSED

        return (name, state)

    def _timer_from_file(self):
        # returns 80% of video duration for the update timer,
        # roughly the time of the video minus the OP and ED
        if self._get_kodi_status() == IDLE:
            return None

        totaltime = self._get_player_props("totaltime")
        seconds = (totaltime['hours'] * 3600) + (totaltime['minutes'] * 60) + (totaltime['seconds'])

        return round(seconds*0.80)

    def _get_rpc_info(self, method, params={}):
        url = "http://{}/jsonrpc".format(self.host_port)

        body = {
            "method": method,
            "params": params,
            "jsonrpc": "2.0",
            "id": 0
        }

        if not self.config['kodi_user'] == '':
            req = urllib.request.Request(url, json.dumps(
                body).encode(), headers=self.headers)
            b64 = base64.b64encode(bytes("{}:{}".format(
                self.config['kodi_user'], self.config['kodi_passwd']), 'ascii'))
            req.add_header("Authorization",
                           "Basic {}".format(b64.decode('utf-8')))
        else:
            req = urllib.request.Request(url, json.dumps(
                body).encode(), headers=self.headers)

        response = urllib.request.urlopen(req)
        data = json.loads(response.read().decode())

        return data['result']

    def _get_player_props(self, prop):
        # Get the required info from the player
        props = {
            "playerid": 1,
            "properties": [prop]
        }

        info = self._get_rpc_info("Player.GetProperties", props)

        return info[prop]

    def observe(self, config, watch_dirs):
        self.msg.info("Using Kodi.")

        while self.active:
            self.status_log.append(self._get_kodi_status())

            if self.status_log[-1] == IDLE and self.status_log[-2] == NOT_RUNNING:
                self.msg.info("Reconnected to Kodi.")

            if self.status_log[-1] == ACTIVE:
                if self.config['kodi_obey_update_wait_s']:
                    self.wait_s = config['tracker_update_wait_s']
                else:
                    self.wait_s = self._timer_from_file()

                player = self._playing_file()
                (state, show_tuple) = self._get_playing_show(player[0])

                self.update_show_if_needed(state, show_tuple)

                if player[1] == PAUSED:
                    self.pause_timer()
                elif player[1] == PLAYING:
                    self.resume_timer()

            elif self.status_log[-1] == AUTH_REQUIRED:
                self.msg.warn("Authentication needed by Kodi, login in the settings and restart trackma.")
            elif self.status_log[-1] == NOT_RUNNING:
                self.msg.warn("Kodi HTTP Server is not running.")

            del self.status_log[0]

            # Wait for the interval before running check again
            time.sleep(config['tracker_interval'])



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# TODO: Add gui stuff for this

import os
import time
import urllib.request
import json

from trackma.tracker import tracker

NOT_RUNNING = 0
ACTIVE = 1
CLAIMED = 2
PLAYING = 3
PAUSED = 4
IDLE = 5


class JellyfinTracker(tracker.TrackerBase):
    name = 'Tracker (Jellyfin)'

    def __init__(self, messenger, tracker_list, config, watch_dirs, redirections=None):
        self.config = config

        self.host_port = self.config['jellyfin_host']+":"+self.config['jellyfin_port']
        self.api_key = self.config['jellyfin_api_key']
        self.username = self.config['jellyfin_user']

        self.status_log = [None, None]

        super().__init__(messenger, tracker_list, config, watch_dirs, redirections)

    def observe(self, config, watch_dirs):
        self.msg.info("Using Jellyfin.")

        while self.active:
            session_info = self._get_sessions_info()
            self.status_log.append(session_info['status'])

            if self.status_log[-1] == ACTIVE or self.status_log[-1] == IDLE:
                if self.status_log[-1] == IDLE and self.status_log[-2] == NOT_RUNNING:
                    self.msg.info("Reconnected to Jellyfin.")
                    self.wait_s = config['tracker_update_wait_s']

                try:
                    (state, show_tuple) = self._get_playing_show(session_info['file_name'])

                    self.view_offset = int(session_info['view_offset'])

                    self.update_show_if_needed(state, show_tuple)

                    if session_info['state'] == PAUSED:
                        self.pause_timer()
                    elif session_info['state'] == PLAYING:
                        self.resume_timer()

                except (IndexError, TypeError):
                    if self.status_log[-1] == IDLE:
                        self.last_filename = None
                        self.update_show_if_needed(0, None)
                    else:
                        pass

            elif self.status_log[-1] == CLAIMED and self.status_log[-2] == CLAIMED:
                self.msg.warn("Claimed Jellyfin, login in the settings and restart trackma.")
            elif self.status_log[-1] == NOT_RUNNING and self.status_log[-2] == NOT_RUNNING:
                self.msg.warn("Jellyfin is not running.")

            del self.status_log[0]

            # Wait for the interval before running check again
            time.sleep(config['tracker_interval'])

    def _get_sessions_info(self):
        session_url = self.host_port+"/Sessions"
        auth_header = "MediaBrowser Token=\"{}\"".format(self.api_key)

        info = {
            "status": NOT_RUNNING,
            "state": PAUSED,
            "file_name": None,
            "view_offset": None
        }

        req = urllib.request.Request(session_url)
        req.add_header("Authorization", auth_header)
        req.add_header("Accept", "application/json")

        with urllib.request.urlopen(req) as response:
            response_json = json.load(response)

        for session in response_json:
            if 'UserName' not in session:
                continue
            if session['UserName'] != self.username:
                continue

            if 'NowPlayingItem' in session:
                current_session = session['NowPlayingItem']
                return {
                    "status": ACTIVE,
                    "state": PAUSED if session['PlayState']['IsPaused'] else PLAYING,
                    "file_name": os.path.basename(current_session['Path']),
                    "view_offset": int(session['PlayState']['PositionTicks']/10000)
                }

        info["status"] = IDLE
        return info



# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import ctypes
import re
import time

from trackma.tracker import tracker


class Win32Tracker(tracker.TrackerBase):
    name = 'Tracker (win32)'

    def __init__(self, messenger, tracker_list, config, watch_dirs, redirections=None):
        super().__init__(messenger, tracker_list, config, watch_dirs, redirections)

        self.winregex = re.compile(r"(\.mkv|\.mp4|\.avi)")

    def _foreach_window(self, hwnd, lParam):
        # Get class name, window title, and process name of the current hwnd
        # and add it to the list of the found windows
        length = ctypes.windll.user32.GetWindowTextLengthW(hwnd)

        pid = ctypes.c_ulong()
        ctypes.windll.user32.GetWindowThreadProcessId(hwnd, ctypes.byref(pid))
        handle = ctypes.windll.kernel32.OpenProcess(0x400 | 0x10, False, pid.value) # PROCESS_QUERY_INFORMATION | PROCESS_VM_READ

        buff_exe = ctypes.create_unicode_buffer(512)
        buff_class = ctypes.create_unicode_buffer(32)
        buff_title = ctypes.create_unicode_buffer(length + 1)

        ctypes.windll.user32.GetClassNameW(hwnd, buff_class, 32)
        ctypes.windll.user32.GetWindowTextW(hwnd, buff_title, length + 1)
        ctypes.windll.psapi.GetModuleBaseNameW(handle, 0, buff_exe, 512)

        ctypes.windll.kernel32.CloseHandle(handle)

        self.win32_hwnd_list.append((buff_class.value, buff_title.value, buff_exe.value))
        return True

    def _get_playing_file_win32(self):
        # Enumerate all windows using the win32 API
        # This will call _foreach_window for each window handle
        # Then return the window title if the class name matches
        # Currently supporting MPC(-HC) and mpv(.net)

        self.win32_hwnd_list = []
        self.EnumWindowsProc = ctypes.WINFUNCTYPE(
            ctypes.c_bool, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))
        ctypes.windll.user32.EnumWindows(
            self.EnumWindowsProc(self._foreach_window), 0)

        for classname, title, processname in self.win32_hwnd_list:
            if classname == 'MediaPlayerClassicW' and self.winregex.search(title) is not None:
                return title
            elif classname == 'mpv' and self.winregex.search(title) is not None:
                return title.replace('mpv - ', '')
            elif processname == 'mpvnet.exe' and self.winregex.search(title) is not None:
                return title.replace(' - mpv.net', '')

        return False

    def observe(self, config, watch_dirs):
        self.msg.info("Using Win32.")

        while self.active:
            # This runs the tracker and update the playing show if necessary
            filename = self._get_playing_file_win32()
            (state, show_tuple) = self._get_playing_show(filename)
            self.update_show_if_needed(state, show_tuple)

            # Wait for the interval before running check again
            time.sleep(1)



# -*- coding: utf-8 -*-
# This file is part of Trackma.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""
Media list provider using Kitsu <https://kitsu.app>.

Example API response: https://kitsu.app/api/edge/anime?filter[text]=build+divide
"""

import datetime
from enum import Enum
import gzip
import json
import socket
import time
import urllib.error
import urllib.parse
import urllib.request

from trackma import utils
from trackma.lib.lib import lib


class PosterImageKey(str, Enum):
    TINY = 'tiny'
    LARGE = 'large'
    SMALL = 'small'
    MEDIUM = 'medium'
    ORIGINAL = 'original'


class libkitsu(lib):
    """
    API class to communicate with Kitsu
    Should inherit a base library interface.

    Website: https://kitsu.app/
    API documentation:
    Designed by:

    """
    name = 'libkitsu'
    user_agent = 'Trackma/{}'.format(utils.VERSION)

    auth = ''
    logged_in = False

    api_info = {
        'name': 'Kitsu',
        'shortname': 'kitsu',
        'version': 'v0.3',
        'merge': True
    }

    default_mediatype = 'anime'
    default_statuses = ['current', 'completed',
                        'on_hold', 'dropped', 'planned']
    default_statuses_dict = {
        'current': 'Watching',
        'completed': 'Completed',
        'on_hold': 'On Hold',
        'dropped': 'Dropped',
        'planned': 'Plan to Watch'
    }

    mediatypes = dict()
    mediatypes['anime'] = {
        'has_progress': True,
        'can_add': True,
        'can_delete': True,
        'can_score': True,
        'can_status': True,
        'can_update': True,
        'can_play': True,
        'statuses_start': ['current'],
        'statuses_finish': ['completed'],
        'statuses_library': ['current', 'on_hold', 'planned'],
        'statuses': default_statuses,
        'statuses_dict': default_statuses_dict,
        'score_max': 5,
        'score_step': 0.25,
    }
    mediatypes['manga'] = {
        'has_progress': True,
        'can_add': True,
        'can_delete': True,
        'can_score': True,
        'can_status': True,
        'can_update': True,
        'can_play': False,
        'statuses_start': ['current'],
        'statuses_finish': ['completed'],
        'statuses': default_statuses,
        'statuses_dict': {
            'current': 'Reading',
            'completed': 'Completed',
            'on_hold': 'On Hold',
            'dropped': 'Dropped',
            'planned': 'Plan to Read'
        },
        'score_max': 5,
        'score_step': 0.25,
    }
    mediatypes['drama'] = {
        'has_progress': True,
        'can_add': True,
        'can_delete': True,
        'can_score': True,
        'can_status': True,
        'can_update': True,
        'can_play': True,
        'statuses_start': ['current'],
        'statuses_finish': ['completed'],
        'statuses': default_statuses,
        'statuses_dict': default_statuses_dict,
        'score_max': 5,
        'score_step': 0.25,
    }

    url = 'https://kitsu.app/api'
    prefix = 'https://kitsu.app/api/edge'

    # TODO : These values are provisional.
    _client_id = 'dd031b32d2f56c990b1425efe6c42ad847e7fe3ab46bf1299f05ecd856bdb7dd'
    _client_secret = '54d7307928f63414defd96399fc31ba847961ceaecef3a5fd93144e960c0e151'

    type_translate = {'ONA': utils.Type.ONA,
                      'OVA': utils.Type.OVA,
                      'TV': utils.Type.TV,
                      'movie': utils.Type.MOVIE,
                      'music': utils.Type.MUSIC,
                      'special': utils.Type.SP}

    status_translate = {'Currently Airing': utils.Status.AIRING,
                        'Finished Airing': utils.Status.FINISHED,
                        'Not Yet Aired': utils.Status.NOTYET}

    def __init__(self, messenger, account, userconfig):
        """Initializes the useragent through credentials."""
        super(libkitsu, self).__init__(messenger, account, userconfig)

        self.username = account['username']
        self.password = account['password']

        # Build opener with the mashape API key
        self.opener = urllib.request.build_opener()
        self.opener.addheaders = [
            ('User-Agent',      self.user_agent),
            ('Accept',          'application/vnd.api+json'),
            ('Accept-Encoding', 'gzip'),
            ('Accept-Charset',  'utf-8'),
        ]

    def _request(self, method, url, get=None, post=None, body=None, auth=False):
        content_type = None

        if get:
            url += "?%s" % urllib.parse.urlencode(get)
        if post:
            post = urllib.parse.urlencode(post).encode('utf-8')
            content_type = 'application/x-www-form-urlencoded'
        if body:
            post = body.encode('utf-8')
            content_type = 'application/vnd.api+json'

        request = urllib.request.Request(url, post)
        request.get_method = lambda: method

        if content_type:
            request.add_header('Content-Type', content_type)

        if auth:
            request.add_header('Authorization', '{0} {1}'.format(
                self._get_userconfig('token_type').capitalize(),
                self._get_userconfig('access_token'),
            ))

        try:
            response = self.opener.open(request)

            # The response most probably will be gzipped so we
            # have to take care of that first
            if response.info().get('content-encoding') == 'gzip':
                return gzip.GzipFile(fileobj=response).read().decode('utf-8')
            else:
                return response.read().decode('utf-8')
        except urllib.error.HTTPError as e:
            if e.code == 401:
                raise utils.APIError("Incorrect credentials.")
            else:
                api_error = self._parse_errors(e)
                if api_error:
                    raise utils.APIError("API error: %s" % api_error)
                else:
                    raise utils.APIError("Connection error: %s" % e)
        except urllib.error.URLError as e:
            raise utils.APIError("URL error: %s" % e)
        except socket.timeout:
            raise utils.APIError("Operation timed out.")

    def _parse_errors(self, e):
        try:
            data = json.loads(e.read().decode('utf-8'))
            errors = ""
            for error in data['errors']:
                errors += "{}: {}".format(error['code'], error['detail'])

            return errors
        except Exception:
            return None

    def _request_access_token(self, refresh=False):
        params = {
            'client_id':     self._client_id,
            'client_secret': self._client_secret,
        }

        if refresh:
            self.msg.info('Refreshing access token...')

            params['grant_type'] = 'refresh_token'
            params['refresh_token'] = self._get_userconfig('refresh_token')
        else:
            self.msg.info('Requesting access token...')

            params['grant_type'] = 'password'
            params['username'] = self.username
            params['password'] = self.password

        response = self._request(
            'POST', self.url + '/oauth/token', post=params)
        data = json.loads(response)

        timestamp = int(time.time())

        self._set_userconfig('access_token',  data['access_token'])
        self._set_userconfig('token_type',    data['token_type'])
        self._set_userconfig('expires',       timestamp + data['expires_in'])
        self._set_userconfig('refresh_token', data['refresh_token'])

        self.logged_in = True
        self._refresh_user_info()
        self._emit_signal('userconfig_changed')

    def _refresh_user_info(self):
        self.msg.info('Refreshing user details...')
        params = {
            "filter[self]": 'true',
        }
        data = self._request('GET', self.prefix +
                             "/users", get=params, auth=True)
        json_data = json.loads(data)
        user = json_data['data'][0]

        # Parse user information
        self._set_userconfig('userid', user['id'])
        self._set_userconfig('username', user['attributes']['name'])

    def check_credentials(self):
        """
        Log into Kitsu. If there isn't an access token, request it, or
        refresh it if necessary.
        """
        timestamp = int(time.time())

        if not self._get_userconfig('access_token'):
            self._request_access_token(False)
        elif (timestamp+60) > self._get_userconfig('expires'):
            self._request_access_token(True)
        else:
            self.logged_in = True

        return True

    def fetch_list(self):
        """Queries the full list from the remote server.
        Returns the list if successful, False otherwise."""
        self.check_credentials()
        self.msg.info('Downloading list...')

        try:
            showlist = dict()
            infolist = list()

            # Get first page and continue from there
            params = {
                "filter[user_id]": self._get_userconfig('userid'),
                "filter[kind]": self.mediatype,
                # "include": self.mediatype, # TODO : This returns a 500 for some reason.
                "include": "media",
                # TODO : List for manga should be different
                f"fields[{self.mediatype}]": ','.join([
                    'id',
                    'slug',
                    'canonicalTitle',
                    'titles',
                    'episodeCount' if self.mediatype in [
                        'anime', 'drama'] else 'chapterCount',
                    'description',
                    'status',
                    'tba',
                    'subtype',
                    'posterImage',
                    'startDate',
                    'endDate',
                    'abbreviatedTitles',
                    'averageRating',
                    'popularityRank',
                    'ratingRank',
                    'ageRating',
                    'ageRatingGuide',
                    'userCount',
                    'favoritesCount'
                ]),
                "page[limit]": "250",
            }

            if self.mediatype == 'anime':
                params['fields[anime]'] += ',nsfw'

            if self.mediatype == 'manga':
                params['fields[manga]'] += ',serialization'

            url = "{}/library-entries?{}".format(
                self.prefix, urllib.parse.urlencode(params))
            i = 1

            while url:
                self.msg.info('Getting page {}...'.format(i))

                data = self._request('GET', url)
                data_json = json.loads(data)

                # print(json.dumps(data_json, sort_keys=True, indent=2))
                # return []

                entries = data_json['data']
                links = data_json['links']

                for entry in entries:
                    # TODO : Including the mediatype returns a 500 for some reason.
                    # showid = int(entry['relationships'][self.mediatype]['data']['id'])
                    showid = int(entry['relationships']['media']['data']['id'])
                    status = entry['attributes']['status']
                    rating = entry['attributes']['ratingTwenty']

                    showlist[showid] = utils.show()
                    showlist[showid].update({
                        'id': showid,
                        'my_id': entry['id'],
                        'my_progress': entry['attributes']['progress'],
                        'my_score': float(rating)/4.00 if rating is not None else 0.0,
                        'my_status': entry['attributes']['status'],
                        'my_start_date': self._iso2date(entry['attributes']['startedAt']),
                        'my_finish_date': self._iso2date(entry['attributes']['finishedAt']),
                        'my_last_update': self._iso2datetime(entry['attributes']['updatedAt']),
                    })

                if 'included' in data_json:
                    medias = data_json['included']
                    for media in medias:
                        info = self._parse_info(media)
                        infolist.append(info)

                    self._emit_signal('show_info_changed', infolist)

                url = links.get('next')
                i += 1

            return showlist
        except urllib.error.HTTPError as e:
            raise utils.APIError(
                "Error getting list (HTTPError): %s" % e.read())
        except urllib.error.URLError as e:
            raise utils.APIError(
                "Error getting list (URLError): %s" % e.reason)

    def merge(self, show, info):
        show['title'] = info['title']
        show['aliases'] = info['aliases']
        show['url'] = info['url']
        show['total'] = info['total']
        show['image'] = info['image']

        show['image_thumb'] = info['image_thumb']

        show['start_date'] = info['start_date']
        show['end_date'] = info['end_date']
        show['status'] = info['status']

    def request_info(self, item_list):
        print("These are missing: " + repr(item_list))
        # TODO implement
        raise NotImplementedError

    def add_show(self, item):
        """Adds a new show in the server"""
        self.check_credentials()
        self.msg.info("Adding show %s..." % item['title'])

        data = self._build_data(item)

        try:
            data = self._request('POST', self.prefix +
                                 "/library-entries", body=data, auth=True)

            data_json = json.loads(data)
            return int(data_json['data']['id'])
        except urllib.error.HTTPError as e:
            raise utils.APIError('Error adding: ' + str(e.code))
        except urllib.error.URLError as e:
            raise utils.APIError('Error adding: ' + str(e.reason))

    def update_show(self, item):
        """Sends a show update to the server"""
        self.check_credentials()
        self.msg.info("Updating show %s..." % item['title'])

        data = self._build_data(item)

        try:
            data = self._request('PATCH', self.prefix + "/library-entries/%s" %
                          item['my_id'], body=data, auth=True)
            data_json = json.loads(data)
            return self._iso2datetime(data_json['data']['attributes']['updatedAt'])
        except urllib.error.HTTPError as e:
            raise utils.APIError('Error updating: ' + str(e.code))
        except urllib.error.URLError as e:
            raise utils.APIError('Error updating: ' + str(e.reason))

    def delete_show(self, item):
        """Sends a show delete to the server"""
        self.check_credentials()
        self.msg.info("Deleting show %s..." % item['title'])

        try:
            self._request('DELETE', self.prefix +
                          "/library-entries/%s" % item['my_id'], auth=True)
        except urllib.error.HTTPError as e:
            raise utils.APIError('Error deleting: ' + str(e.code))
        except urllib.error.URLError as e:
            raise utils.APIError('Error deleting: ' + str(e.reason))

    def search(self, query, method):
        self.msg.info("Searching for %s..." % query)

        values = {
            "filter[text]": query,
            "page[limit]": 20,
        }

        try:
            data = self._request('GET', self.prefix +
                                 "/" + self.mediatype, get=values)
            shows = json.loads(data)

            infolist = []
            for media in shows['data']:
                info = self._parse_info(media)
                infolist.append(info)

            self._emit_signal('show_info_changed', infolist)

            if not infolist:
                raise utils.APIError('No results.')

            return infolist
        except urllib.error.HTTPError as e:
            raise utils.APIError('Error searching: ' + str(e.code))
        except urllib.error.URLError as e:
            raise utils.APIError('Error searching: ' + str(e.reason))

    def _build_data(self, item):
        values = {'data': {
            'type': 'libraryEntries',
            'attributes': {},
            'relationships': {
                'media': {
                    'data': {
                        'type': self.mediatype,
                        'id': item['id'],
                    }
                },
                'user': {
                    'data': {
                        'type': 'users',
                        'id': self._get_userconfig('userid'),
                    }
                },
            }
        }
        }

        # Update necessary keys
        if item['my_id']:
            values['data']['id'] = str(item['my_id'])
        if 'my_progress' in item:
            values['data']['attributes']['progress'] = item['my_progress']
        if 'my_status' in item:
            values['data']['attributes']['status'] = item['my_status']
        if 'my_score' in item:
            values['data']['attributes']['ratingTwenty'] = int(
                item['my_score']*4) or None

        return json.dumps(values)

    def _str2date(self, string):
        if string is None:
            return None

        try:
            return datetime.datetime.strptime(string, "%Y-%m-%d")
        except Exception:
            self.msg.debug('Invalid date {}'.format(string))
            return None  # Ignore date if it's invalid

    def _iso2date(self, string):
        if string is None:
            return None

        try:
            return datetime.datetime.strptime(string, "%Y-%m-%dT%H:%M:%S.%fZ").date()
        except Exception:
            self.msg.debug('Invalid date {}'.format(string))
            return None  # Ignore date if it's invalid

    def _iso2datetime(self, string):
        if string is None:
            return None

        try:
            return datetime.datetime.strptime(string, "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=datetime.timezone.utc)
        except Exception:
            self.msg.debug('Invalid date {}'.format(string))
            return None  # Ignore date if it's invalid

    def _guess_status(self, start_date, end_date):
        # Try to guess show status by checking start and end dates
        now = datetime.datetime.now()

        if end_date and end_date < now:
            return utils.Status.FINISHED

        if start_date:
            if start_date > now:
                return utils.Status.NOTYET
            else:
                return utils.Status.AIRING

        # Safe to assume dates haven't even been announced yet
        return utils.Status.NOTYET

    def _parse_info(self, media):
        info = utils.show()
        attr = media['attributes']

        if media['type'] == 'anime':
            total = attr['episodeCount']
        elif media['type'] == 'manga':
            total = attr['chapterCount']
        elif media['type'] == 'drama':
            total = attr['episodeCount']

        poster_image = attr.get('posterImage', {})
        image = utils.get_any(poster_image, PosterImageKey.SMALL, PosterImageKey.MEDIUM, PosterImageKey.LARGE, PosterImageKey.ORIGINAL)
        image_thumb = utils.get_any(poster_image, PosterImageKey.TINY)

        info.update({
            'id': int(media['id']),
            # TODO : Some shows actually don't have a canonicalTitle; this should be fixed in the future.
            # For now I'm just picking the romaji title in these cases.
            'title':       attr['titles'].get('en_jp') or attr.get('canonicalTitle') or attr['titles'].get('en'),
            'total':       total or 0,
            'image':       image,
            'image_thumb': image_thumb,
            'start_date':  self._str2date(attr['startDate']),
            'end_date':    self._str2date(attr['endDate']),
            'type':        self.type_translate.get(attr['subtype'], utils.Type.UNKNOWN),
            'status':      self.status_translate.get(attr['status'], utils.Status.UNKNOWN),
            'url': "https://kitsu.app/{}/{}".format(self.mediatype, attr['slug']),
            'aliases':     list(filter(None, attr['titles'].values())),
            'extra': [
                ('Synopsis',            attr['description']),
                ('Type',                attr['subtype']),
                ('Titles',              list(
                    filter(None, attr['titles'].values()))),
                ('Synonyms',            attr.get('abbreviatedTitles', '?')),
                ('Average Rating',      attr.get('averageRating', '?')),
                ('Rank',                "Popularity #{}\nRating #{}".format(
                    attr.get('popularityRank', '?'), attr.get('ratingRank', '?'))),
                ('Age Rating',          "{} ({})".format(
                    attr.get('ageRating', 'Unknown'), attr.get('ageRatingGuide', 'Unknown'))),
            ]
        })

        if attr.get('nsfw', False):
            info['extra'].insert(
                2, ('NSFW', 'Yes'))
        elif attr.get('serialization', False):
            info['extra'].insert(
                4, ('Serialization', attr['serialization']))

        # WORKAROUND: Shows with 1 episode (TVs, SPs, OVAs) end the same day they start
        if total == 1:
            info['end_date'] = info['start_date']

        if attr['status'] in ['upcoming', 'unreleased']:
            info['extra'].append(('Expected Release', attr.get('tba', '?')))

        return info
